# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT License, included in this distribution as LICENSE.txt

""" """

from rowgenerators.source import Source

class FixedSource(Source):
    """Generate rows from a fixed-width source"""

    def __init__(self, ref, table=None, cache=None, working_dir=None, env=None, **kwargs):
        super().__init__(ref, cache, working_dir, **kwargs)

        self.table = table

        assert self.table

    def __iter__(self):
        """Iterate over all of the lines in the file"""

        self.start()

        parse = self.table.make_fw_row_parser()


        with open(self.ref.fspath) as f:
            for line in f.readlines():
                yield parse(line)

        self.finish()




# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT License, included in this distribution as LICENSE.txt

""" """

from rowgenerators.exceptions import RowGeneratorError
from rowgenerators.source import Source
from xlrd import open_workbook, XLRDError
from rowgenerators.util import md5_file

class ExcelSource(Source):
    """Generate rows from an excel file"""

    def __init__(self, ref, cache=None, working_dir=None, env=None, **kwargs):
        super().__init__(ref, cache, working_dir, **kwargs)

        self.url = ref

        # It is supposed to be segment. Or file. Probably segment. Well, one of them.
        # ts = self.url.target_segment or self.url.target_file
        # if not ts:
        #    raise RowGeneratorError("URL does not include target file in fragment: {}".format(self.url))

    @staticmethod
    def srow_to_list(row_num, s):
        """Convert a sheet row to a list"""

        values = []

        try:
            for col in range(s.ncols):
                values.append(s.cell(row_num, col).value)
        except:
            raise

        return values

    def __iter__(self):
        """Iterate over all of the lines in the file"""

        self.start()

        wb = open_workbook(filename=str(self.url.fspath))

        ts = self.url.target_segment or  0

        # Without this check, failure to provide a target_segment will cause the return
        # of the first worksheet.

        # if not ts:
        #    raise RowGeneratorError("URL does not include target file in fragment: {}".format(self.url))

        try:
            try:
                sheets = wb.sheets()
                sheet_no = int(ts)
                s = sheets[sheet_no]
            except (ValueError, IndexError):
                # ValueError when Segment is the workbook name, not the number
                # IndexError when the segment is a numeric name, such as a year,
                # which converted with int(), but is larger than the # of sheets
                s = wb.sheet_by_name(ts)

        except XLRDError as e:
            raise RowGeneratorError("Failed to open Excel workbook: '{}' ".format(e))

        for i in range(0, s.nrows):
            yield self.srow_to_list(i, s)

        self.finish()

    @property
    def children(self):
        """Return the sheet names from the workbook """
        from xlrd import open_workbook

        wb = open_workbook(filename=str(self.url.fspath))

        sheets = wb.sheet_names()

        return sheets

    @staticmethod
    def make_excel_date_caster(file_name):
        """Make a date caster function that can convert dates from a particular workbook. This is required
        because dates in Excel workbooks are stupid. """

        from xlrd import open_workbook

        wb = open_workbook(file_name)
        datemode = wb.datemode

        def excel_date(v):
            from xlrd import xldate_as_tuple
            import datetime

            try:

                year, month, day, hour, minute, second = xldate_as_tuple(float(v), datemode)
                return datetime.date(year, month, day)
            except ValueError:
                # Could be actually a string, not a float. Because Excel dates are completely broken.
                from dateutil import parser

                try:
                    return parser.parse(v).date()
                except ValueError:
                    return None

        return excel_date



# Copyright (c) 2018 Civic Knowledge. This file is licensed under the terms of the
# MIT, included in this distribution as LICENSE

"""Appurls or sql databases"""

from rowgenerators.appurl import Url
from rowgenerators.appurl.file import FileUrl


class Sql(Url):
    """"""

    match_priority = Url.match_priority - 1

    def __init__(self, url=None, downloader=None, **kwargs):
        # Save for auth_url()

        self.username

        self._orig_url = url
        self._orig_kwargs = dict(kwargs.items())

        super().__init__(url, downloader=downloader, **kwargs)

    def get_resource(self):
        return self

    def get_target(self):
        return self

    @property
    def dsn(self):
        """Return a database connection string. The string will have values interpolated from
        the environment"""
        from rowgenerators.exceptions import RowGeneratorError
        import os

        u = self.clone()
        u.fragment = []
        try:
            u.password = u.password.format(**os.environ)
        except KeyError as e:
            raise RowGeneratorError(("Failed to set password from environment variable in connection string '{}'. "
                                     " exception = {}").format(str(u), str(e)))

        return str(u)

    @property
    def sql(self):
        """Return the query, which is embedded in the fragment"""
        return self._fragment[0]

    @sql.setter
    def sql(self, value):
        """Return the query, which is embedded in the fragment"""
        self._fragment[0] = value

class SqlFile(Sql):
    """
    Files that hold a single SQL Select, and end with the file extension .sql
    """
    @classmethod
    def _match(cls, url, **kwargs):
        return url.target_format == 'sql'

class InlineSqlUrl(Sql):
    """
    SQL Urls of the Form:

        sql://dsn#<SQL Query>

    The dsn component must be resolved to another Sql URL, external to this system.

    """

    def __init__(self, url=None, downloader=None, dsns=None, **kwargs):
        super().__init__(url, downloader, **kwargs)

        self.dsns = dsns or {}  # A dict of DSNs for DSN resolution

    def get_resource(self):
        from rowgenerators.appurl.url import parse_app_url
        from rowgenerators.exceptions import AppUrlError

        if self.dsn_name not in self.dsns:
            raise AppUrlError(
                ("Resolving a SqlDsn URL requires a DSN dict with the name '{}' in it. ".format(self.dsn_name) +
                 "Pass the dict to parse_app_url as parse_app_url('urlstr', dsns={<dsn_dict>}) "))

        u = parse_app_url(self.dsns[self.dsn_name])

        u._fragment[0] = self.sql

        return u

    @classmethod
    def _match(cls, url, **kwargs):
        return url.proto == 'sql'

    @property
    def dsn_name(self):
        """Return the name of the DSN that should be used"""

        return self.hostname


class OracleSql(Sql):
    generator_class = Sql

    @classmethod
    def _match(cls, url, **kwargs):
        return url.proto == 'oracle'



# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT License, included in this distribution as LICENSE.txt

""" """

from .csv import CsvSource

class TsvSource(CsvSource):
    """Generate rows from a TSV (tab separated value) source"""

    delimiter = '\t'


class PipeSource(CsvSource):
    """Generate rows from a file delimited by a pipe character"""

    delimiter = '|'





"""Math functions available for use in derivedfrom columns

Copyright (c) 2015 Civic Knowledge. This file is licensed under the terms of
the Revised BSD License, included in this distribution as LICENSE.txt

"""

import datetime
import dateutil.parser as dp

from .measures import IntValue, FloatValue, DateValue



class NullValue(Exception):
    """Raised from a caster to indicate that the returned value should be None"""


def transform_generator(fn):
    """A decorator that marks transform pipes that should be called to create the real transform"""
    fn.__dict__['is_transform_generator'] = True
    return fn


def is_transform_generator(fn):
    """Return true of the function has been marked with @transform_generator"""
    try:
        return fn.__dict__.get('is_transform_generator', False)
    except AttributeError:
        return False


def row_number(row_n):
    return row_n


def nullify(v):
    """Convert empty strings and strings with only spaces to None values. """

    if isinstance(v, str):
        v = v.strip()

    if v is None or v == '':
        return None
    else:
        return v


def clean_float(v):
    """Remove commas from a float"""

    if v is None or not str(v).strip():
        return None

    return float(str(v).replace(',', '').replace(' ',''))


def clean_int(v):
    """Remove commas from a float"""

    if v is None or not str(v).strip():
        return None

    return int(str(v).replace(',', '').replace(' ',''))


#
# Casters that return a default value
#
def int_d(v, default=None):
    """Cast to int, or on failure, return a default Value"""

    try:
        return int(v)
    except:
        return default


def float_d(v, default=None):
    """Cast to int, or on failure, return a default Value"""

    try:
        return float(v)
    except:
        return default


#
# Casters that return a null on failure
#

def int_n(v):
    """Cast to int, or on failure, return a default Value"""

    try:
        return int(float(v))
    except:
        return None


def float_n(v):
    """Cast to float, or on failure, return None"""

    try:
        return float(v)  # Just to be sure the code property exists
    except:
        return None


def int_e(v):
    """Cast to int, or on failure raise a NullValue exception"""

    try:
        return int(v)
    except:
        raise NullValue(v)


def parse_int(v, header_d):
    """Parse as an integer, or a subclass of Int."""
    from rowgenerators.rowpipe.exceptions import CastingError

    v = nullify(v)

    if v is None:
        return None

    try:
        # The converson to float allows converting float strings to ints.
        # The conversion int('2.134') will fail.
        return int(round(float(v), 0))
    except (TypeError, ValueError) as e:
        raise CastingError(int, header_d, v, 'Failed to cast to integer')


def parse_float(v, header_d):
    from rowgenerators.rowpipe.exceptions import CastingError

    v = nullify(v)

    if v is None:
        return None

    try:
        return float(v)
    except (TypeError, ValueError) as e:
        raise CastingError(float, header_d, v, str(e))


def parse_str(v, header_d):
    # TODO: It's so complicated while py2/py3 work because str is binary for py2, but unicode for py3.

    # This is often a no-op, but it ocassionally converts numbers into strings

    v = nullify(v)

    if v is None:
        return None

    return _parse_text(v, header_d)


def parse_bytes(v, header_d):
    return _parse_binary(v, header_d)


def parse_unicode(v, header_d):
    return _parse_text(v, header_d)


def parse_type(type_, v, header_d):
    from rowgenerators.rowpipe.exceptions import CastingError

    v = nullify(v)

    if v is None:
        return None

    try:
        return type_(v)
    except (TypeError, ValueError) as e:
        raise CastingError(type_, header_d, v, str(e))


def parse_date(v, header_d):
    from rowgenerators.rowpipe.exceptions import CastingError

    v = nullify(v)

    if v is None:
        return None

    if isinstance(v, str):
        try:
            return dp.parse(v).date()
        except (ValueError, TypeError) as e:
            raise CastingError(datetime.date, header_d, v, str(e))

    elif isinstance(v, datetime.date):
        return v
    else:
        raise CastingError(int, header_d, v, "Expected datetime.date or basestring, got '{}'".format(type(v)))


def parse_time(v, header_d):
    from rowgenerators.rowpipe.exceptions import CastingError

    v = nullify(v)

    if v is None:
        return None

    if isinstance(v, str):
        try:
            return dp.parse(v).time()
        except ValueError as e:
            raise CastingError(datetime.time, header_d, v, str(e))

    elif isinstance(v, datetime.time):
        return v
    else:
        raise CastingError(int, header_d, v, "Expected datetime.time or basestring, got '{}'".format(type(v)))


def parse_datetime(v, header_d):
    from rowgenerators.rowpipe.exceptions import CastingError

    v = nullify(v)

    if v is None:
        return None

    if isinstance(v, str):
        try:
            return dp.parse(v)
        except (ValueError, TypeError) as e:
            raise CastingError(datetime.datetim, header_d, v, str(e))

    elif isinstance(v, datetime.datetime):
        return v
    else:
        raise CastingError(int, header_d, v, "Expected datetime.time or basestring, got '{}'".format(type(v)))


class IntOrNone(IntValue):
    "An Integer value that stores values that fail to convert in the 'code' property"
    _pythontype = int

    code = None

    def __new__(cls, v):
        try:
            o = super(IntOrCode, cls).__new__(cls, v)
        except Exception as e:
            o = super(IntOrCode, cls).__new__(cls, None)
            o.code = v
        return o


class FloatOrNone(FloatValue):
    "An Float value that stores values that fail to convert in the 'code' property"
    _pythontype = float

    code = None

    def __new__(cls, v):
        try:
            o = super(FloatOrCode, cls).__new__(cls, v)
        except Exception as e:
            o = super(FloatOrCode, cls).__new__(cls, None)
            o.code = v
        return o


class IntOrCode(IntValue):
    "An Integer value that stores values that fail to convert in the 'code' property"
    _pythontype = int

    code = None

    def __new__(cls, v):
        try:
            o = super(IntOrCode, cls).__new__(cls, v)
        except Exception as e:
            o = super(IntOrCode, cls).__new__(cls, 0)
            o.code = v
        return o


class FloatOrCode(FloatValue):
    "An Float value that stores values that fail to convert in the 'code' property"
    _pythontype = float

    code = None

    def __new__(cls, v):
        try:
            o = super(FloatOrCode, cls).__new__(cls, v)
        except Exception as e:
            o = super(FloatOrCode, cls).__new__(cls, float('nan'))
            o.code = v
        return o


class DateOrCode(DateValue):
    "An Integer value that stores values that fail to convert in the 'code' property"
    _pythontype = datetime.date

    code = None

    def __new__(cls, v):
        try:
            o = super(DateOrCode, cls).__new__(cls, v)
        except Exception as e:
            o = super(DateOrCode, cls).__new__(cls, None)
            o.code = v
        return o


class ForeignKey(IntValue):
    """An Integer value represents a foreign key on another table.  The value can hold a linked row for access from other
    columns. """

    _pythontype = int

    row = None

    def __new__(cls, v):
        o = super(ForeignKey, cls).__new__(cls, v)
        return o

    def __init__(self, v):
        super(ForeignKey, self).__init__()
        self.row = None


def _parse_text(v, header_d):
    """ Parses unicode.

    Note:
        unicode types for py2 and str types for py3.

    """

    from rowgenerators.rowpipe.exceptions import CastingError

    v = nullify(v)

    if v is None:
        return None

    try:
        return str(v).strip()
    except Exception as e:
        raise CastingError(str, header_d, v, str(e))


def _parse_binary(v, header_d):
    """ Parses binary string.

    Note:
        <str> for py2 and <binary> for py3.

    """

    # This is often a no-op, but it ocassionally converts numbers into strings

    v = nullify(v)

    if v is None:
        return None

    try:
        return bytes(v, 'utf-8').strip()
    except UnicodeEncodeError:
        return str(v).strip()


def excel_dt_1900(v):
    """Convert a float that representes a date in an excel file into a datetime. The float
    is assumed to have a basis of 1900"""

    from xlrd.xldate import xldate_as_datetime
    return xldate_as_datetime(v, 0)


def excel_dt_1904(v):
    """Convert a float that representes a date in an excel file into a datetime. The float
    is assumed to have a basis of 1904"""

    from xlrd.xldate import xldate_as_datetime
    return xldate_as_datetime(v, 0)


#def date(v):
#    """Convert a date to a datetime"""
#    return v.date()



# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# Revised BSD License, included in this distribution as LICENSE

"""
CLI program for managing Metatab files
"""




def appurl():
    import sys
    from pkg_resources import iter_entry_points
    from tabulate import tabulate
    from rowgenerators.appurl import parse_app_url

    import argparse
    parser = argparse.ArgumentParser(
        prog='appurl',
        description='app url configuration program',
       )

    g = parser.add_mutually_exclusive_group(required=True)

    g.add_argument('-l', '--list', action='store_true',
                   help="List all of the registered AppUrl handlers")

    g.add_argument('-i', '--info', help="INformatino about a URL. May download resources")

    args = parser.parse_args(sys.argv[1:])

    entries = []
    for ep in iter_entry_points('appurl.urls'):
        c = ep.load()
        entries.append([c.match_priority, ep.name, ep.module_name,  c.__name__, ])


    if args.list:
        print(tabulate(sorted(entries), ['Priority', 'EP Name', 'Module', 'Class'] ))
    elif args.info:

        u = parse_app_url(args.info)
        r = u.get_resource()
        t = r.get_target()

        t = [
            ('Url', str(u)),
            ('Url Type', type(u)),
            ('Generator Type', type(u.generator)),
            ('Resource', str(r)),
            ('Target', str(t))

        ]

        print(tabulate(t))

if __name__ == "__main__":
    # execute only if run as a script
    appurl()


from __future__ import print_function

import unittest
import warnings

from rowgenerators.rowpipe import RowProcessor
from rowgenerators.rowpipe import Table

warnings.resetwarnings()
warnings.simplefilter("ignore")


class TestBasic(unittest.TestCase):

    def setUp(self):
        import warnings
        super().setUp()

        warnings.simplefilter('ignore')

    def test_table(self):

        from rowgenerators.rowpipe import Table

        t = Table('foobar')
        t.add_column('i1', datatype='int')
        t.add_column('i2', valuetype='int')
        t.add_column('i3', valuetype='measure/int')
        t.add_column('f1', datatype='float')
        t.add_column('f2', valuetype='float')
        t.add_column('f3', valuetype='measure/float')

        self.assertEqual(6, len(list(t)))

        for c in t:
            print(c)

    def test_expand_transform_1(self):
        from rowgenerators.rowpipe import Table
        from rowgenerators.rowpipe import RowProcessor

        def doubleit(v):
            return int(v) * 2

        env = {
            'doubleit': doubleit
        }

        t = Table('extable')
        t.add_column('id', datatype='int')
        t.add_column('b', datatype='int')
        t.add_column('v1', datatype='int', transform='^row.a')
        t.add_column('v2', datatype='int', transform='row.v1;doubleit')
        t.add_column('v3', datatype='int', transform='^row.a;doubleit')

        for c in t:
            print('---', c)
            for i, tr in enumerate(c.expanded_transform):
                print('   ', i, len(list(tr)), list(tr))

        headers = ['stage'] + list(c.name for c in t)

        table = [[i] + [tr.str(i) for tr in stage] for i, stage in enumerate(t.stage_transforms)]

        from tabulate import tabulate

        print(tabulate(table, headers, tablefmt="rst"))

        class Source(object):

            headers = 'a b'.split()

            def __iter__(self):
                for i in range(N):
                    yield i, 2 * i

        rp = RowProcessor(Source(), t, env=env, code_path='/tmp/rowgenerators/test_transform.py')

    def test_expand_transform_2(self):
        from rowgenerators.rowpipe import Table
        from rowgenerators.rowpipe import RowProcessor

        def doubleit(v):
            return int(v) * 2

        env = {
            'doubleit': doubleit
        }

        t = Table('extable')
        t.add_column('id', datatype='int')
        t.add_column('v4', datatype='float', transform='^row.a;doubleit;doubleit')
        t.add_column('v5', datatype='int', transform='^row.a;doubleit|doubleit')
        t.add_column('v6', datatype='str', transform="^str('v6-string')")

        for c in t:
            print('---', c)
            for i, tr in enumerate(c.expanded_transform):
                print('   ', i, len(list(tr)), list(tr))

        headers = ['stage'] + list(c.name for c in t)

        table = [[i] + [tr.str(i) for tr in stage] for i, stage in enumerate(t.stage_transforms)]

        from tabulate import tabulate

        print(tabulate(table, headers, tablefmt="rst"))

        class Source(object):

            headers = 'a b'.split()

            def __iter__(self):
                for i in range(N):
                    yield i, 2 * i

        rp = RowProcessor(Source(), t, env=env, code_path='/tmp/rowgenerators/test_transform.py')

    # NOTE. This speed test is about 12x to 23x faster running in PyPy than CPython!
    def test_basic_transform(self):

        from rowgenerators.rowpipe import Table
        from rowgenerators.rowpipe import RowProcessor
        from contexttimer import Timer

        def doubleit(v):
            return int(v) * 2

        env = {
            'doubleit': doubleit
        }

        t = Table('foobar')
        t.add_column('id', datatype='int')
        t.add_column('a', datatype='int')
        t.add_column('v1', datatype='int', transform='^row.a')
        t.add_column('v2', datatype='int', transform='row.v1;doubleit')
        t.add_column('v3', datatype='int', transform='^row.a;doubleit')
        t.add_column('v4', datatype='float', transform='^row.a;doubleit;doubleit')
        t.add_column('v5', datatype='int', transform='^row.a;doubleit|doubleit')
        t.add_column('v6', datatype='float')

        N = 20000

        class Source(object):

            headers = 'a b'.split()

            def __iter__(self):
                for i in range(N):
                    yield i, 2 * i

        rp = RowProcessor(Source(), t, env=env, code_path='/tmp/rowgenerators/test_transform.py')

        print("Code: ", rp.code_path)

        headers = rp.headers

        for row in rp:
            d = dict(zip(headers, row))

            self.assertEqual(d['a'], d['v1'], d)
            self.assertEqual(2 * d['a'], d['v2'], d)
            self.assertEqual(2 * d['a'], d['v3'], d)
            self.assertEqual(4 * d['a'], d['v4'], d)
            self.assertEqual(4 * d['a'], d['v5'], d)

        count = 0
        row_sum = 0
        with Timer() as t:
            for row in rp:
                count += 1

                row_sum += round(sum(row[:6]))

        self.assertEqual(2199890000, row_sum)

        print('Rate=', float(N) / t.elapsed)

    def test_init_transform(self):

        from rowgenerators.rowpipe import Table

        def expand_transform(code, datatype='int', valuetype=None):
            t = Table('foobar')
            c = t.add_column('c', datatype=datatype, valuetype=valuetype, transform=code)
            return c.expanded_transform

        print(expand_transform('^GeoidCensusTract|v.as_acs()'))
        print(expand_transform('v.as_acs()'))

    def test_many_transform(self):

        from rowgenerators.rowpipe import Table
        from rowgenerators.rowpipe import RowProcessor
        from contexttimer import Timer

        def doubleit(v):
            return int(v) * 2

        def printstuff(v, manager, accumulator):
            print(type(accumulator), accumulator)
            return v

        def accumulate(v, accumulator):
            from collections import deque

            if not 'deque' in accumulator:
                accumulator['deque'] = deque([0], 3)

            accumulator['deque'].append(v)

            return sum(accumulator['deque'])

        def addtwo(x):
            return x + 2

        env = {
            'doubleit': doubleit,
            'printstuff': printstuff,
            'accumulate': accumulate,
            'addtwo': addtwo
        }

        transforms = [
            ('', 45),
            ('^row.a', 45),
            ('^row.a;doubleit', 90),
            ('^row.a;doubleit;doubleit', 180),
            ('^row.a;doubleit|doubleit', 180),
            ('^row.a;row.c*2|doubleit', 180),
            ('^row.a;row.c/3|doubleit', 24),
            ('doubleit', 90),
            ('doubleit;doubleit', 180),
            ('doubleit|doubleit', 180),
            ('row.c*2|doubleit', 180),
            ('row.c/3|doubleit', 24),
            ('accumulate', 109),
            ('manager.factor_a*row.c', 450),
            ('addtwo(row.c)', 65),
        ]

        N = 10

        class Source(object):
            headers = 'a'.split()

            def __iter__(self):
                for i in range(N):
                    yield (i,)

        class Manager(object):
            factor_a = 10

        for i, (tr, final_sum) in enumerate(transforms):
            t = Table('foobar')
            t.add_column('c', datatype='int', transform=tr)
            rp = RowProcessor(Source(), t, env=env,
                              manager=Manager(),
                              code_path='/tmp/rowgenerators/test_many_transform_{}.py'.format(i))

            row_sum = 0
            with Timer() as t:
                for row in rp:
                    row_sum += sum(row)

            self.assertEqual(final_sum, row_sum)

    def test_source_dest_diff(self):
        from tabulate import tabulate

        class Source(object):

            headers = 'Z X Y'.split()

            def __iter__(self):
                yield self.headers
                for i in range(5):
                    yield ('Z'+str(i), 'X'+str(i),  'Y'+str(i))

        t = Table('extable')
        t.add_column('a', datatype='int')
        t.add_column('b', datatype='int')
        t.add_column('c', datatype='int')

        source_headers = 'c b a'.split()

        rp = RowProcessor(Source(), t, source_headers=source_headers,
                          code_path='/tmp/rowgenerators/test_transform.py')

        print(tabulate(rp))

        print(rp.source_headers)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Civic Knowledge. This file is licensed under the terms of the
# MIT License, included in this distribution as LICENSE.txt

"""

"""

from .file import FileUrl
from rowgenerators.appurl.web import WebUrl

from rowgenerators.appurl.archive.zip import ZipUrl

class ShapefileUrl(ZipUrl):

    match_priority = ZipUrl.match_priority - 1

    def __init__(self, url=None, downloader=None, **kwargs):
        super().__init__(url, downloader, **kwargs)

        self.scheme_extension = 'shape'

        if not self.target_file:
            self.target_file = '.*\.shp$'

    @classmethod
    def _match(cls, url, **kwargs):
        return url.scheme_extension == 'shape'

    def get_resource(self):
        return super().get_resource()

    def list(self):
        """List the files in the referenced Zip file"""

        from zipfile import ZipFile

        real_files = ZipUrl.real_files_in_zf(ZipFile(self.fspath))
        return list(self.set_target_file(rf) for rf in real_files)

    def get_target(self):
        """Returns the ZIP file, with the correct fragment, not the inner
        file as a ZipUrl normally does. """

        if not str(self.fspath).endswith('.shp'):
            #Resolve the target_file, which may be a reg-ex
            self.target_file = ZipUrl.get_file_from_zip(self)

        return self

    def get_archive_target(self):
        """Returns the inner .shp file, like a Normal ZipUrl would"""

        self.fragment = [ZipUrl.get_file_from_zip(self), self.fragment[1]]

        t = super().get_target().clear_fragment()

        return ShapefileShpUrl(str(t))

    def geoframe(self, *args, **kwargs):

        import geopandas

        return self.get_resource().get_target().generator.geoframe( *args, **kwargs)

class ShapefileShpUrl(FileUrl):

    match_priority = 1000 # Don't match this one.

    def __init__(self, url=None, downloader=None, **kwargs):
        super().__init__(url, downloader, **kwargs)

        self.scheme_extension = 'shape'

        self.fragment = None

    @classmethod
    def _match(cls, url, **kwargs):
        return False


    def get_resource(self):
        return self

    def get_target(self):
        """Returns the ZIP file, with the correct fragmentment, not the inner
        file as a ZipUrl normally does. """

        return self

    def get_archive_target(self):
        """Returns the inner .shp file, like a Normal ZipUrl would"""

        return self


#! /usr/bin/env python -u

from string import ascii_uppercase
import sys
import csv
from os import getenv, environ
import json
import select

if __name__ == "__main__":

    try:
        w = csv.writer(sys.stdout)

        w.writerow(['row','type','k','v'])

        i=0

        for k in sys.argv:
            w.writerow([i,'arg', k, ''])
            i += 1

        for k, v in environ.items():
            w.writerow([i,'env',k,v])
            i += 1

        for k, v in json.loads(environ.get('PROPERTIES')).items():
            w.writerow([i, 'prop', k, v])
            i += 1
    except BrokenPipeError:
        # Parent exited before we wrote everything. This is almost guaranteed to
        # happen while building schemas with `metapack -s`
        pass




# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT License, included in this distribution as LICENSE.txt

""" """

import unittest
from os.path import dirname
import sys

from rowgenerators import get_generator, parse_app_url
from rowgenerators.generator.csv import CsvSource
from rowgenerators.test import get_file, data_path, script_path, RowGeneratorTest


class BasicTests(RowGeneratorTest):

    def test_csv(self):

        us = 'http://public.source.civicknowledge.com/example.com/sources/simple-example-altnames.csv'

        self.assertEqual(10001, len(list(CsvSource(get_file(us)))))

        us = 'http://public.source.civicknowledge.com/example.com/sources/unicode-utf8.csv'

        self.assertEqual(53, len(list(CsvSource(get_file(us)))))

    def test_entrypoints(self):
        from rowgenerators.generator.iterator import IteratorSource
        from rowgenerators.generator.generator import GeneratorSource
        from rowgenerators.generator.csv import CsvSource

        us = 'http://public.source.civicknowledge.com/example.com/sources/unicode-utf8.csv'

        def g():
            yield None

        self.assertIsInstance(get_generator([]), IteratorSource)
        self.assertIsInstance(get_generator(g()), GeneratorSource)
        self.assertIsInstance(get_generator(parse_app_url(us).get_resource().get_target()), CsvSource)

    def test_sources(self):
        from csv import DictReader

        with open(data_path('sources.csv')) as f:
            for e in DictReader(f):

                if not e['url_class']:
                    print()
                    continue

                u = parse_app_url(e['url'])
                r = u.get_resource()
                t = r.get_target()

                g = get_generator(t)

                self.assertEquals(e['gen_class'], g.__class__.__name__)

                self.assertEquals(int(e['n_rows']), (len(list(g))))

    @unittest.skip('Broken') #skipIf(sys.platform.startswith("win"), "requires Windows")
    def test_program(self):

        u = parse_app_url(script_path('rowgen.py'))
        u.scheme_extension = 'program'

        env = {
            '--long-arg': 'a',
            '-s': 'a',
            'ENV_VAR':'a',
            'prop1':'a',
            'prop2':'a'
        }

        g = get_generator(u, working_dir=dirname(u.path), env=env)

        print(type(g))

        rows = {}

        for row in g.iter_rp:
            rows[row['type']+'-'+row['k']] = row.v
            print(row)

        self.assertEqual('a', rows['prop-prop1'])
        # SHould be self.assertEqual('{"prop1": "a", "prop2": "a"}', rows['env-PROPERTIES'])
        # but there are sorting issues, and it's a string, not an actual dict
        self.assertIn('"prop1": "a"', rows['env-PROPERTIES'])
        self.assertIn('"prop2": "a"', rows['env-PROPERTIES'])


    @unittest.skip('Has local path')
    def test_fixed(self):
        from itertools import islice

        from rowgenerators.table import  Table, Column

        t = Table()
        t.add_column('id',int,6)
        t.add_column('uuid', str, 34)
        t.add_column('int', int, 3)
        t.add_column('float', float, 14)

        print(str(t))

        parse = t.make_fw_row_parser()

        u = parse_app_url('fixed+file:/Volumes/Storage/Downloads/test_data/fixed/simple-example.txt')

        print(u.get_resource())
        print(u.get_resource().get_target())

        g = get_generator(u, table=t)

        print(type(g))

        for row in islice(g,10):
            print(row)

    def test_google(self):

        from hashlib import md5
        from rowgenerators.appurl.web.google import GoogleSpreadsheetUrl

        url = 'gs:1qjjtkMqpxtkDp3qZlkF7P8Tm8VtfIwiWW-OqJ2J91yE#2038675149'

        u = parse_app_url(url)
        self.assertIsInstance(u, GoogleSpreadsheetUrl)

        wu = u.web_url

        r = u.get_resource()

        t = r.get_target()

        h = md5()

        for r in t.generator:
            h.update( (''.join(e for e in r)).encode('utf8'))

        self.assertEqual(h.hexdigest(), '2108f203561d6b4cf3b6ecf74bd769c6')

        u = parse_app_url('https://docs.google.com/spreadsheets/u/0/d/e/2PACX-1vS6_JK5zktVQr6JwkYUPvzlwcw0YAawSVC7ldWZVfg9hvTjBxl2z4xWaWCrzb9JZ0Go07KhLgbzw5DW/pubhtml/sheet?headers=false&gid=1658358543')

        self.assertIsInstance(u, GoogleSpreadsheetUrl)


    def test_RowGenerator(self):
        import warnings
        warnings.simplefilter("ignore")

        from rowgenerators import RowGenerator

        rg = RowGenerator('http://public.source.civicknowledge.com/example.com/sources/simple-example-altnames.csv')

        self.assertEqual(10001, len(list(rg)))

        df = rg.dataframe()

        self.assertEqual(10000, len(df))

        url = 'http://public.source.civicknowledge.com/example.com/sources/simple-example.foo#&target_format=csv'

        self.assertEqual(10001, len(list(RowGenerator(url))))

        url = 'http://public.source.civicknowledge.com/example.com/sources/simple-example.foo'

        self.assertEqual(10001, len(list(RowGenerator(url, target_format='csv' ))))

    def test_api(self):

        import  rowgenerators as rg

        df = rg.dataframe('http://public.source.civicknowledge.com/example.com/sources/simple-example-altnames.csv',
                          na_filter=True)
        self.assertEqual(496703.2204, df.baz.astype(float).sum().round(4))

        gdf = rg.geoframe('http://library.metatab.org/census.gov-tracts-2017-sandiego-1/data/tracts.csv')
        self.assertEqual(41.6682, gdf.area.sum().round(4))

    @unittest.skip('Has local path')
    def test_h5(self):

        import h5py
        import pandas as pd
        from rowgenerators.appurl.file.hdf5 import Hdf5Url
        from rowgenerators.generator.hdf5 import Hdf5Source

        #fn_base = data_path('small_demo.h5')+'#nlsy97_all_1997-2013;'
        fn_base = '/Users/eric/proj/virt-proj/data-project/sdrdl-data-projects/nlsinfo.org/' \
                  'lsy97_all_1997-2013/nlsy97_all_1997-2013.h5'+'#nlsy97_all_1997-2013;'
        fn_slice = '1,2,3,4'
        fn = fn_base+fn_slice

        u = parse_app_url(fn)

        self.assertIsInstance(u,Hdf5Url)
        self.assertIsInstance(u.generator, Hdf5Source)

        df = u.dataframe()
        self.assertEqual(27,df.head().sum().sum())
        self.assertEqual(['B0000300', 'B0000400', 'B0000500', 'B0000600'],list(df.columns))

        fn = fn_base + ','.join(df.columns)

        u = parse_app_url(fn)

        df = u.dataframe()
        self.assertEqual(27, df.head().sum().sum())
        self.assertEqual(['B0000300', 'B0000400', 'B0000500', 'B0000600'], list(df.columns))





# Copyright (c) 2016 Civic Knowledge. This file is licensed under the terms of the
# MIT License, included in this distribution as LICENSE.txt

"""Pipes, pipe segments and piplines, for flowing data from sources to partitions.

"""

import textwrap
import os
import sys

class RowPipeError(Exception):
    pass

class RowProcessorError(Exception):
    pass


class CasterExceptionError(RowPipeError):

    def __init__(self,  function, field_header, value,  exc, exec_info, *args, **kwargs):

        exc_type, exc_obj, exc_tb = exec_info

        fname = exc_tb.tb_frame.f_code.co_filename
        linen = exc_tb.tb_lineno

        message = "Failed to cast column '{}' value='{}' in function='{}', file='{}', line='{}' : {} "\
            .format(field_header, value,  function,
                    fname, linen,
                    str(exc))

        # Call the base class constructor with the parameters it needs
        Exception.__init__(self, textwrap.fill(message, 120), *args, **kwargs)


class CastingError(RowPipeError):

    def __init__(self, type_target, field_header, value, message, *args, **kwargs):

        self.type_target = type_target
        self.field_header = field_header
        self.value = value

        message = "Failed to cast column '{}' value='{}' to '{}': {} "\
            .format(field_header, value, type_target, message)

        # Call the base class constructor with the parameters it needs
        Exception.__init__(self, textwrap.fill(message, 120), *args, **kwargs)


class TooManyCastingErrors(RowPipeError):

    def __init__(self, *args, errors = None, **kwargs):
        self.errors = errors if errors is not None else {}
        Exception.__init__(self, *args, **kwargs)







# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT, included in this distribution as LICENSE

""" """




def path2url(path):
    "Convert a pathname to a file URL"

    from urllib.parse import urljoin
    from urllib.request import pathname2url


    return urljoin('file:', pathname2url(path))

def parse_file_to_uri(url):
    """If this is a file path, return a Path object, otherwise, return None"""
    import pathlib
    from urllib.parse import urlparse,urlunparse
    import re

    url = str(url)

    p = urlparse(url)

    if p.scheme == 'file':
        return url

    if p.fragment:
       url = re.sub('\#[^#]*', '', url)

    try:
        if re.match(r'^[a-zA-Z]:', url): # Has a drive letter
            uri = pathlib.PureWindowsPath(url.replace('\\','/')).as_uri()

        elif url.startswith('//') or url.startswith('\\\\'): # Windows share
            uri = pathlib.PureWindowsPath(url.replace('\\', '/')).as_uri()

        elif not p.scheme: # No scheme, but can't use .as_uri if relative
            uri = pathlib.PurePath(url.replace('\\', '/')).as_uri()
        else:
            uri = None

    except ValueError as e: #ValueError: relative path can't be expressed as a file URI
        uri = 'file:'+p.path

    if uri:
        if p.fragment:
            uri = str(uri) + ('#'+p.fragment)

        return uri

    else:
        return None


def parse_url_to_dict(url, assume_localhost=False):
    """Parse a url and return a dict with keys for all of the parts.

    The urlparse function() returns a wacky combination of a namedtuple
    with properties.

    """

    import re
    from urllib.parse import unquote_plus, ParseResult, urlparse, parse_qs

    assert url is not None

    url = str(parse_file_to_uri(url) or url)

    p = urlparse(url)

    #  '+' indicates that the scheme has a scheme extension
    if '+' in p.scheme:
        scheme_extension, scheme = p.scheme.split('+')
    else:
        scheme = p.scheme
        scheme_extension = None

    if scheme == '':
        # If we continue from here, just setting the scheme,
        # fragments won't parse correctly, but they do if the scheme is properly set
        scheme = 'file'
        #return parse_url_to_dict('file://'+)

    frag_parts = (unquote_plus(p.fragment) if p.fragment else '').split('&')

    if frag_parts and '=' not in frag_parts[0]:
        # Maybe the first part are the segments, one or two names seperated by ';'
        frag = frag_parts.pop(0)
    else:
        frag = None

    frag_query = dict(tuple(e.split('=')) for e in frag_parts )

    if frag:
        frag_sub_parts = frag.split(';')

        if len(frag_sub_parts) < 2:
            frag_sub_parts = [frag_sub_parts[0], None]

    else:
        frag_sub_parts = [None, None]

    # Urlparse converts the hostname to lowercase, which we'd prefer it not do when the
    # hostname is specified as an interpoation string.
    # Note that this will only work if the hostname is entirely uppercase, and will fail if
    # it is mixed case.

    if p.hostname and (p.hostname[0] == '{' and p.hostname[-1] == '}' and p.hostname.upper() in url):
        p_hostname = p.hostname.upper()
    else:
        p_hostname = p.hostname

    def unmangle_windows_path(scheme, path):
        import re
        if scheme == 'file' and re.match("/[a-zA-Z]:", path):
            return path.lstrip('/')
        else:
            return path

    return {
        'scheme': scheme,
        'scheme_extension': scheme_extension,
        'netloc': p.netloc,
        'hostname': p_hostname,
        'path': unmangle_windows_path(scheme,p.path),
        'params': p.params,
        'query': p.query,
        'fragment':  frag_sub_parts,
        'fragment_query': frag_query,
        'username': p.username,
        'password': p.password,
        'port': p.port
    }

def unparse_fragment(d):

    from urllib.parse import quote_plus, urlencode, unquote

    if d.get('fragment') or d.get('fragment_query'):

        if isinstance(d.get('fragment'),(list, tuple)):

            seg = ';'.join(quote_plus(str(e)) for e in [ e for e in d.get('fragment') if e])
        else:
            seg = quote_plus(d.get('fragment'))

        if d.get('fragment_query'):
            fqt = sorted(d.get('fragment_query').items())
            query = '&' + urlencode(fqt,doseq=True)
        else:
            query = ''


        if seg or query:
            return "#"+seg+unquote(query)

    return ''

def unparse_url_dict(d, **kwargs):
    from urllib.parse import quote_plus, urlencode, unquote
    import re

    d = dict(d.items())

    d.update(kwargs)

    if '_fragment' in d and 'fragment' not in d:
        d['fragment'] = d['_fragment']


    # Using netloc preserves case for the host, which the host value does not do,
    # but netloc also will have the username, password and port in it
    if d.get('netloc') and ('@' not in d['netloc'] and ':' not in d['netloc']):
        if 'netloc' in d and d['netloc']:
            host_port = d['netloc']
        else:
            host_port = ''
    else:
        host_port = d['hostname']

    if 'port' in d and d['port']:
        host_port += ':' + str(d['port'])

    user_pass = ''
    if 'username' in d and d['username']:
        user_pass += d['username']

    if 'password' in d and d['password']:
        user_pass += ':' + d['password']

    if user_pass:
        host_port = '{}@{}'.format(user_pass, host_port)



    if d.get('scheme') == 'file':
        if d['netloc']: # Windows UNC path
            url = 'file://{}/{}'.format(d['netloc'], d.get('path', '').lstrip('/'))

        elif d.get('path','').startswith('/'): # absolute path
            url = 'file://{}'.format( d.get('path', ''))

        elif re.match('[a-zA-Z]:/', d.get('path','')): # windows path
            url = 'file:///{}'.format(d.get('path', ''))

        else: # relative path
            url = 'file:{}'.format( d.get('path', ''))

    elif d.get('scheme') == 'mailto':
        url = '{}:{}'.format(d['scheme'], d.get('path', ''))

    elif d.get('scheme') and host_port:
        url = '{}://{}/{}'.format(d['scheme'], host_port, d.get('path', '').lstrip('/'))

    elif d.get('scheme'):
        url = '{}:{}'.format(d['scheme'], d.get('path', '').lstrip('/'))

    elif d.get('path'):
        raise AppUrlError("Can only unparse file urls that have a file: scheme")
        # It's possibly just a local file url.
        # This isn't the standard file: url form, which is specified to have a :// and a host part,
        # like 'file://localhost/etc/config', but that form can't handle relative URLs ( which don't start with '/')
        # url = 'file:'+d['path']
    else:
        url = ''

    if d.get('scheme_extension'):
        url = d['scheme_extension']+'+'+url

    if 'query' in d and d['query']:
        url += '?' + d['query']

    url = url + unparse_fragment(d)

    return url

def reparse_url(url, **kwargs):

    assume_localhost = kwargs.get('assume_localhost', False)

    return unparse_url_dict(parse_url_to_dict(url,assume_localhost),**kwargs)

def join_url_path(url, *paths):
    """Like path.os.join, but operates on the url path, ignoring the query and fragments."""

    import os

    parts = parse_url_to_dict(url)

    return reparse_url(url, path=os.path.join(parts['path']))

def file_ext(v):
    """Split of the extension of a filename, without throwing an exception of there is no extension. Does not
    return the leading '.'
    :param v: """

    from os.path import splitext

    if not v:
        return None

    try:
        try:
            v = splitext(v)[1][1:]
        except TypeError:
            v = splitext(v.fspath)[1][1:]

        if v == '*':  # Not a file name, probably a fragment regex
            return None

        return v.lower() if v else None
    except IndexError:
        return None



def copy_file_or_flo(input_, output, buffer_size=64 * 1024, cb=None):
    """ Copy a file name or file-like-object to another file name or file-like object"""

    from os import makedirs
    from os.path import isdir, dirname

    assert bool(input_)
    assert bool(output)

    input_opened = False
    output_opened = False

    try:
        if isinstance(input_, str):

            if not isdir(dirname(input_)):
                makedirs(dirname(input_))

            input_ = open(input_, 'r')
            input_opened = True

        if isinstance(output, str):

            if not isdir(dirname(output)):
                makedirs(dirname(output))

            output = open(output, 'wb')
            output_opened = True

        # shutil.copyfileobj(input_,  output, buffer_size)

        def copyfileobj(fsrc, fdst, length=buffer_size):
            cumulative = 0
            while True:
                buf = fsrc.read(length)
                if not buf:
                    break
                fdst.write(buf)
                if cb:
                    cumulative += len(buf)
                    cb(len(buf), len(buf), cumulative)

        copyfileobj(input_, output)

    finally:
        if input_opened:
            input_.close()

        if output_opened:
            output.close()

DEFAULT_CACHE_NAME = 'rowgen-cache'

def get_cache_name(cache_name=None):
    cn = cache_name or DEFAULT_CACHE_NAME

    if not cn:
        from rowgenerators.exceptions import ConfigurationError

        raise ConfigurationError("Must either set the default cache name or call cache functions with a name")

    return cn

def set_default_cache_name(cache_name):
    global DEFAULT_CACHE_NAME
    DEFAULT_CACHE_NAME = cache_name

def get_cache(cache_name=None, clean=False):
    """Return the path to a file cache"""

    from fs.osfs import OSFS
    from fs.appfs import UserDataFS
    from fs.errors import CreateFailed
    import os

    cache_name = get_cache_name(cache_name)

    # If the environmental variable for the cache is set, change the cache directory.

    env_var = cache_name.upper().replace('-','_')

    if 'CACHE' not in env_var:
        env_var += '_CACHE'

    cache_dir = os.getenv(env_var, None)

    if cache_dir:
        try:
            r =  OSFS(cache_dir)
        except CreateFailed as e:
            raise CreateFailed("Failed to create '{}': {} ".format(cache_dir, e))
    else:

        try:
            r = UserDataFS(cache_name.lower())
        except CreateFailed as e:
            raise CreateFailed("Failed to create '{}': {} ".format(cache_name.lower(), e))


    return r


def clean_cache(cache = None, cache_name=None):
    """Delete items in the cache older than 24 hours"""
    import datetime

    cache = cache if cache else get_cache( get_cache_name(cache_name))

    ignores = ['index.json', 'index.json.bak']

    for step in cache.walk.info():
        details = cache.getdetails(step[0])
        mod = details.modified
        now = datetime.datetime.now(tz=mod.tzinfo)
        age = (now - mod).total_seconds()
        if age > (60 * 60 * 24) and details.is_file:
            if step[0] not in ignores:
                cache.remove(step[0])

def nuke_cache(cache = None, cache_name=None):
    """Delete Everythong in the cache"""

    cache = cache if cache else get_cache(get_cache_name(cache_name))

    for step in cache.walk.info():
        if not step[1].is_dir:
            cache.remove(step[0])

def ensure_dir(path):
    from os import makedirs
    from os.path import exists

    if path and not exists(path):
            makedirs(path)


def import_name_or_class(name):
    " Import an object as either a fully qualified, dotted name, "

    if isinstance(name, str):

        # for "a.b.c.d" -> [ 'a.b.c', 'd' ]
        module_name, object_name = name.rsplit('.',1)

        # __import__ loads the multi-level of module, but returns
        # the top level, which we have to descend into
        mod = __import__(module_name)

        components = name.split('.')

        for comp in components[1:]: # Already got the top level, so start at 1
            mod = getattr(mod, comp)

        return mod
    else:
        return name # Assume it is already the thing we want to import




# Copyright (c) 2016 Civic Knowledge. This file is licensed under the terms of the
# MIT License, included in this distribution as LICENSE.txt

"""
Tables and columns

"""

from itertools import zip_longest

from rowgenerators.exceptions import ConfigurationError
from rowgenerators.exceptions import SchemaError
from rowgenerators.table import Column as RGColumn
from rowgenerators.table import Table as RGTable
from rowgenerators.valuetype import resolve_value_type


class Table(RGTable):

    def add_column(self, name, datatype=None, valuetype=None, transform=None, width=None):
        c = Column(name, datatype, width, valuetype, transform)
        self.columns.append(c)
        return c

    @property
    def stage_transforms(self):
        """Expanded transforms, organized as stages. Each entry is a list of one stage of transforms for all columns.
        Each stage has an entry for every colum, even when no transform is specified for that column, either
        setting the datatype for the first stage, or a passthrough function for any others."""

        stages = list(zip_longest(*[c.expanded_transform for c in self]))

        new_stages = []

        columns = list(self)

        for i, stage in enumerate(stages):
            new_stage = []
            for j, tr in enumerate(stage):
                if tr:
                    new_stage.append(tr)
                elif i == 0:
                    new_stage.append(TransformSegment(datatype=columns[j].valuetype or columns[j].datatype,
                                                      column=columns[j]))
                else:
                    new_stage.append(TransformSegment(transforms=['v'], column=columns[j]))

            new_stages.append(new_stage)

        return new_stages

    def __str__(self):
        from tabulate import tabulate
        headers = 'name datatype valuetype transform width'.split()
        rows = [(c.name, c.datatype.__name__, c.valuetype.__name__, c.transform, c.width) for c in self.columns]

        return ('Table: {}\n'.format(self.name)) + tabulate(rows, headers)


class TransformSegment(object):

    def __init__(self, init=None, datatype=None, transforms=None, exception=None, column=None):
        self.init = init
        self.transforms = transforms or []
        self.datatype = datatype
        self.exception = exception
        self.column = column

    def __getitem__(self, item):
        return getattr(self, item)

    def __setitem__(self, key, value):
        return setattr(self, key, value)

    def __iter__(self):
        """Iterate the sequence of parts, including the init, datatype and transforms"""

        if self.init:
            yield self.init

        if self.datatype:
            yield self.datatype

        yield from self.transforms

    def __repr__(self):
        fields = []
        for f in 'init datatype transforms exception'.split():
            if self[f]:
                if f == 'datatype':
                    fields.append('{}={}'.format(f, self[f].__name__))
                else:
                    fields.append('{}={}'.format(f, self[f]))

        return "<Transform {} {} >".format(self.column.name, ' '.join(fields))

    def str(self, stage_n):

        return '|'.join(e.__name__ if isinstance(e, type) else e for e in list(self))


class Column(RGColumn):

    def __init__(self, name, datatype=None, width=None, valuetype=None, transform=None):

        if valuetype is not None and datatype is None:
            self.valuetype = resolve_value_type(valuetype)

        elif datatype is not None:
            self.valuetype = resolve_value_type(datatype)
        else:
            self.valuetype = None

        if self.valuetype is None:
            raise SchemaError("Could not resovle type for for column '{}' datatype='{}' and valuetype='{}' "
                              .format(name, datatype, valuetype))

        self.transform = transform

        if width is not None and width != '':
            width = int(width)

        super().__init__(name, self.valuetype.python_type(), width)

    def __repr__(self):
        return "<Column {name} dt={datatype} vt={valuetype} {transform}>" \
            .format(name=self.name, datatype=self.datatype.__name__, valuetype=self.valuetype.__name__,
                    transform=self.transform)

    @property
    def dict(self):
        return dict(
            name=self.name,
            datatype=self.datatype,
            valuetype=self.valuetype,
            transform=self.transform
        )

    @property
    def expanded_transform(self):
        """Expands the transform string into segments """

        transform = self.transform.rstrip('|') if self.transform else ''

        segments = [TransformSegment(column=self)]

        for i, seg_str in enumerate(transform.split(';')):  # ';' seperates pipe stages

            pipes = seg_str.split('|')  # seperate pipes in each stage.

            d = TransformSegment(column=self)

            for pipe in pipes:

                if not pipe.strip():
                    continue

                if pipe[0] == '^':  # First, the initializer
                    if segments[0].init:
                        raise ConfigurationError('Can only have one initializer in a pipeline segment')
                    if i != 0:
                        raise ConfigurationError('Can only have an initializer in the first pipeline segment')
                    segments[0].init = pipe[1:]  # initializers only go on the first segment

                elif pipe[0] == '!':  # Exception Handler
                    d.exception = pipe[1:]
                else:  # Assume before the datatype
                    d['transforms'].append(pipe.strip())

            if d['transforms'] or d.exception:
                segments.append(d)

        if not segments[0].init:
            segments[0].init = self.valuetype or self.datatype

        return segments




# From https://stackoverflow.com/a/3431838
def md5(fname):
    import hashlib
    hash_md5 = hashlib.md5()
    with open(fname, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()

def data_path(v):
    from os.path import dirname, join
    d = dirname(__file__)
    return join(d, 'test_data', v)


def script_path(v=None):
    from os.path import dirname, join
    d = dirname(__file__)
    if v is not None:
        return join(d, 'scripts', v)
    else:
        return join(d, 'scripts')

def sources():
    import csv
    with open(data_path('sources.csv')) as f:
        r = csv.DictReader(f)
        return list(r)


def cache_fs():
    from fs.tempfs import TempFS

    return TempFS('rowgenerator')

# From https://stackoverflow.com/a/3431838
def md5(fname):
    import hashlib
    hash_md5 = hashlib.md5()
    with open(fname, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()






"""


Copyright (c) 2015 Civic Knowledge. This file is licensed under the terms of
the Revised BSD License, included in this distribution as LICENSE.txt

"""

from .core import *

class ErrorVT(ValueType):
    role = ROLE.ERROR
    vt_code = 'e'

    def __init__(self,v):
        pass

class MarginOfErrorVT(FloatValue):
    role = ROLE.ERROR
    vt_code = 'm'

    def __init__(self,v):
        pass

    @property
    def se(self):

        if '/90' in self.vt_code:
            return self / 1.645
        elif '/95' in self.vt_code:
            return self / 1.96
        elif '/99' in self.vt_code:
            return self / 2.575
        else:
            raise ValueError("Can't understand value_args: {}".format(self.type_args))

    @property
    def m90(self):
        return self.se * 1.645

    @property
    def m95(self):
        return self.se * 1.96

    @property
    def m99(self):
        return self.se * 2.575

    def ci90u(self, v):
        """Return the 90% confidence interval upper value"""
        return v + self.m90

    def ci90l(self, v):
        """Return the 90% confidence interval lower value"""
        return v - self.m90

    @property
    def m95(self):
        pass

    def ci90u(self, v):
        """Return the 95% confidence interval upper value"""
        return v + self.m90

    def ci90l(self, v):
        """Return the 95% confidence interval lower value"""
        return v - self.m90

    def ci95u(self, v):
        """Return the 95% confidence interval upper value"""
        return v + self.m95

    def ci95l(self, v):
        """Return the 95% confidence interval lower value"""
        return v - self.m95


    def rse(self, denom):
        """Return the relative standard error, relative to the given value"""
        return self.se / denom


class ConfidenceIntervalHalfVT(FloatValue):
    """An upper or lower half of a confidence interval"""
    role = ROLE.ERROR
    vt_code = 'ci'

    def __init__(self,v):
        pass

    @property
    def se(self):
        pass


class StandardErrorVT(FloatValue):
    """A standard error"""
    role = ROLE.ERROR
    vt_code = 'se'

    @property
    def m90(self):
        o =  MarginOfErrorVT(self * 1.645)
        o.vt_code += '/90'
        return o

    @property
    def m95(self):
        o = MarginOfErrorVT(self * 1.96)
        o.vt_code += '/95'
        return o

    @property
    def m99(self):
        o = MarginOfErrorVT(self * 2.575)
        o.vt_code += '/99'
        return o

    def ci90u(self, v):
        """Return the 95% confidence interval upper value"""
        return v + self.m90

    def ci90l(self, v):
        """Return the 95% confidence interval lower value"""
        return v - self.m90

    def ci95u(self, v):
        """Return the 95% confidence interval upper value"""
        return v + self.m95

    def ci95l(self, v):
        """Return the 95% confidence interval lower value"""
        return v - self.m95


class RelativeStandardErrorVT(FloatValue):
    role = ROLE.ERROR
    vt_code = 'rse'


error_value_types = {
    "margin": MarginOfErrorVT, # m90, m95, m99
    "ci": ConfidenceIntervalHalfVT, #ci90u, ci90l, ci95u, ci95l, ci99u, ci99l
    "se": StandardErrorVT,
    "rse": RelativeStandardErrorVT,
}


# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT, included in this distribution as LICENSE

""" """

import logging
from functools import lru_cache

_logger = logger = logging.getLogger('rowgenerators.appurl.web.download')


class _NoOpFileLock(object):
    """No Op for pyfilesystem caches where locking wont work"""

    def __init__(self, lf):
        pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_val:
            raise exc_val

    def acquire(self):
        pass

    def release(self):
        pass


class Resource(object):
    cache_path = None
    sys_path = None
    download_time = None

    def __init__(self):
        super().__init__()

    def __str__(self):
        return str(self.__dict__)

def default_downloader_callback(msg_type, message, read_len, total_len):
    pass


class Downloader(object):
    """Downloader objects handle downloading resrouces from the web, including authorization,
    and storing the downloaded object in a cache. Since they are the primary interface to the file cache,
    all Urls object have a link to a Downloader """

    context = {}  # A variable substitution context, for substituting hostnames, pathnames, etc

    singleton = None

    def __init__(self, cache=None, account_accessor=None, logger=None,
                 working_dir='', callback=None, use_cache=True):
        """
        Download and cache files, via HTTP and FTP, with retry and decompression.

        :param self:
        :param cache: A PyFs filesystem object for caching files
        :param account_accessor: An objevct for acessing account credentials. Not currently used.
        :param logger: Logging object to write debug logs to
        :param working_dir:
        :param callback: Call back to call with progress reports during downloads.
        :return:
        """

        self._cache = cache
        self.account_acessor = account_accessor
        self.logger = logger
        self.working_dir = working_dir
        self._callback = callback or default_downloader_callback

        self.use_cache = use_cache # Set to false to ignore cache

        # For debugging singletonness
        #from metapack.util import dump_stack
        #print('======')
        #print(dump_stack(5))

    @staticmethod
    def get_instance(cache=None, account_accessor=None, logger=None,
                     working_dir='', callback=None):
        """Return a memoized singleton"""

        # Wont see this debug statement with the --debug cli command ; it gets called
        # before setting debug log_level
        if not Downloader.singleton:
            _logger.debug("Creating downloader for cache '{}' ".format(cache))
            Downloader.set_singleton(cache, account_accessor, logger, working_dir, callback)

        return Downloader.singleton

    @staticmethod
    def set_singleton(cache=None, account_accessor=None, logger=None,
                     working_dir='', callback=None):
        """Assign the downloader singleton"""

        Downloader.singleton = Downloader(cache, account_accessor, logger, working_dir, callback)
        return Downloader.singleton

    def callback(self, msg_type, message, read_len=-1, total_len=-1 ):
        if True or self._callback:
            self._callback( msg_type, message, read_len, total_len)

    def set_callback(self, cb):
        # The Downloader is supposed to be a singleton, but it gets created all over the place
        self._callback = cb

    def reset_callback(self):
        # The Downloader is supposed to be a singleton, but it gets created all over the place
        self.set_callback(default_downloader_callback)



    @property
    def cache(self):
        if not self._cache:
            from rowgenerators import get_cache
            # qn = self.__module__+'.'+self.__class__.__qualname__
            self._cache = get_cache()

        return self._cache

    def get_resource(url):
        pass

    def download(self, url):
        from os.path import abspath, join
        from genericpath import exists

        from rowgenerators.appurl.url import Url
        from rowgenerators.exceptions import DownloadError, AccessError

        logger.debug(f"Download {url}")

        working_dir = self.working_dir if self.working_dir else ''

        r = Resource()

        # For local files, don't download, just reference in place.
        if url.scheme == 'file':
            r.cache_path = Url(url.resource_url).path
            r.download_time = None

            # Many places the file may exist
            locations = {  # What a mess ...
                abspath(r.cache_path),
                abspath(r.cache_path.lstrip('/')),
                abspath(join(working_dir, r.cache_path)),
                abspath(r.cache_path.lstrip('/'))
            }

            for l in locations:
                if exists(l):
                    r.sys_path = l
                    logger.debug("Found '{}'as local file '{}'".format(str(url), l))
                    break
            else:
                raise DownloadError(("File resource does not exist. Found none of:"
                                     "\n{}\n\nWorking dir = {}\ncache_path={}\nspec_path={}")
                                    .format('\n'.join(locations), working_dir, r.cache_path, url.path))

        else:
            # Not a local file, so actually need to download it.
            try:
                r.cache_path, r.download_time = self._download_with_lock(url.resource_url)
            except AccessError as e:
                # Try again, using a URL that we may have configured an account for. This is
                # primarily S3 urls, with Boto or AWS credential
                try:
                    r.cache_path, r.download_time = self._download_with_lock(url.auth_resource_url)
                except AttributeError as e:
                    raise e
                except DownloadError as e:
                    raise DownloadError("Access error for url '{}'; also tried accessing as S3 url '{}'".format(url, url.auth_resource_url ))

            r.sys_path = self.cache.getsyspath(r.cache_path)

        logger.debug(f"Downloaded {url} to {r.cache_path}")

        return r

    def cache_path(self, url):
        import hashlib
        from urllib.parse import urlparse
        from os import sep
        from os.path import join, dirname, basename

        url = url.replace('\\', '/')

        # .decode('utf8'). The fs modulegets upset when given strings, so
        # we need to decode to unicode. UTF8 is a WAG.
        try:
            parsed = urlparse(url.decode('utf8'))
        except AttributeError:
            parsed = urlparse(url)

        # Create a name for the file in the cache, based on the URL
        # the '\' replacement is because pyfs only wants to use UNIX path seperators, but
         # python os.path.join will use the one specified for the operating system.
        cache_path = join(parsed.netloc, parsed.path.strip('/'))

        # If there is a query, hash it and add it to the path
        if parsed.query:
            hash = hashlib.sha224(parsed.query.encode('utf8')).hexdigest()
            # We put the hash before the last path element, because that's the target faile, which gets
            # used to figure out what the target format should be.
            cache_path = join(dirname(cache_path), hash, basename(cache_path))

        cache_path  = cache_path.replace(sep,'/')

        return cache_path

    def _download_with_lock(self, url):
        """
        Download a URL and store it in the cache.

        :param url:
        :param cache_fs:
        :param account_accessor: callable of one argument (url) returning dict with credentials.
        :param clean: Remove files from cache and re-download
        :param logger:
        :param callback:
        :return:
        """

        import os.path
        from os.path import join
        import time

        from fs.errors import DirectoryExpected, NoSysPath, ResourceInvalid, DirectoryExists
        from requests import HTTPError
        from rowgenerators.exceptions import AccessError, DownloadError

        assert isinstance(url, str)

        cache_path = self.cache_path(url)

        if not self.cache.exists(cache_path):

            cache_dir = os.path.dirname(cache_path)

            try:
                self.cache.makedirs(cache_dir, recreate=True)
            except DirectoryExpected as e:

                # Probably b/c the dir name is already a file
                dn = os.path.dirname(cache_path)
                bn = os.path.basename(cache_path)
                for i in range(10):
                    try:
                        cache_path = join(dn + str(i), bn)
                        self.cache.makedirs(os.path.dirname(cache_path))
                        break
                    except DirectoryExpected as e2:
                        continue
                    except DirectoryExists:
                        pass  # ? No idea what's supposed to happen here.
                    raise e
                else:
                    # Exhausted all of the trial values
                    raise e

        try:
            from filelock import FileLock
            lock = FileLock(self.cache.getsyspath(cache_path + '.lock'))

        except NoSysPath:
            # mem: caches, and others, don't have sys paths.
            # FIXME should check for MP operation and raise if there would be
            # contention. Mem  caches are only for testing with single processes
            lock = _NoOpFileLock()

        with lock:
            if self.cache.exists(cache_path):
                # Rather than ignoring the cache when use_cache is False, we
                # delete the file and re-download it, because if you ignore the cached file,
                # you still have to download the resource to a file somewhere.
                if self.use_cache:
                    logger.debug(f"Found {cache_path} in cache, and cache is active")
                    return cache_path, None
                else:
                    logger.debug(f"File {cache_path} is not in cache")
                    try:
                        logger.debug(f"Found {cache_path} in cache, but cache not active; deleting")
                        self.cache.remove(cache_path)
                    except ResourceInvalid:
                        pass  # Well, we tried.

            try:
                self._download(url, cache_path)

                return cache_path, time.time()

            except HTTPError as e:
                if e.response.status_code == 403:
                    raise AccessError("Access error on download: {}".format(e))
                else:
                    raise DownloadError("Failed to download: {}".format(e))

            except (KeyboardInterrupt, Exception):
                # This is really important -- its really bad to have partly downloaded
                # files being confused with fully downloaded ones.
                # FIXME. Should also handle signals. deleting partly downloaded files is important.
                # Maybe should have a sentinel file, or download to another name and move the
                # file after done.
                if self.cache.exists(cache_path):
                    self.cache.remove(cache_path)

                raise

        assert False, 'Should never get here'

    def _download(self, url, cache_path):
        import requests
        import functools
        from urllib.request import urlopen

        from requests.exceptions import SSLError

        from rowgenerators.appurl.util import parse_url_to_dict, copy_file_or_flo
        from rowgenerators.exceptions import DownloadError
        from ftplib import FTP

        self.callback('download', url)

        if url.startswith('s3:'):

            from rowgenerators import parse_app_url

            s3url = parse_app_url(url)

            try:
                with self.cache.open(cache_path, 'wb') as f:
                    s3url.object.download_fileobj(f)
            except Exception as e:
                raise DownloadError("Failed to fetch S3 url '{}': {}".format(url, e))

        elif url.startswith('ftp:'):

            logger.debug("Fetch " + str(url))

            u = parse_url_to_dict(url)

            try:
                with FTP(u['netloc']) as ftp, self.cache.open(cache_path, 'wb') as fout:

                    total_len = [0]

                    def _read(d):
                        fout.write(d)

                        total_len[0] = total_len[0] + len(d)

                        self.callback('ftp read', url, len(d), total_len[0])

                    ftp.login()
                    ftp.retrbinary('RETR ' + u['path'], _read)
                    ftp.quit()
            except ConnectionError as e:
                raise DownloadError("Failed to get FTP url '{}': {} ".format(url, e))

        else:

            logger.debug("Request " + str(url))

            try:
                r = requests.get(url, stream=True)
                r.raise_for_status()
            except SSLError as e:
                raise DownloadError("Failed to GET {}: {} ".format(url, e))

            if r.status_code>=300:
                if r.status_code>=304:
                    raise DownloadError("Server Returned 304: Not Modified");
                else:
                    raise DownloadError("Can't handle server response, {r.status_code}");

            # Requests will auto decode gzip responses, but not when streaming. This following
            # monkey patch is recommended by a core developer at
            # https://github.com/kennethreitz/requests/issues/2155
            if r.headers.get('content-encoding') == 'gzip':
                r.raw.read = functools.partial(r.raw.read, decode_content=True)

            def copy_cb(message, read_len, total_len):
                # Message is just the read len
                self.callback('copy', url, read_len, total_len)

            logger.debug(f"Response code={r.status_code} {r.headers}" )

            with self.cache.open(cache_path, 'wb') as f:
                copy_file_or_flo(r.raw, f, cb=copy_cb)

            assert self.cache.exists(cache_path)




# Copyright (c) 2016 Civic Knowledge. This file is licensed under the terms of the
# Revised BSD License, included in this distribution as LICENSE

from __future__ import print_function

import sys
from itertools import islice

from pkg_resources import iter_entry_points
from rowgenerators.appurl.enumerate import enumerate_contents
from rowgenerators.appurl.url import Url
from rowgenerators.exceptions import SourceError, TextEncodingError
from rowgenerators.source import Source
# Change the row cache name
from rowgenerators.util import get_cache
from tableintuit import RowIntuiter
from tabulate import tabulate


def prt(*args):
    print(*args)


def warn(*args):
    print('WARN:', *args)


def err(*args):
    import sys
    print("ERROR:", *args)
    sys.exit(1)


def run_row_intuit(path, cache):
    for encoding in ('ascii', 'utf8', 'latin1'):
        try:
            rows = list(islice(Source(url=path, encoding=encoding, cache=cache), 5000))
            return encoding, RowIntuiter().run(rows)
        except TextEncodingError:
            pass

    raise Exception('Failed to convert with any encoding')


def rowgen():
    import argparse

    parser = argparse.ArgumentParser(
        prog='rowgen',
        description='Return CSV rows of data from a rowgenerator URL')

    parser.add_argument('-H', '--head', default=False, action='store_true',
                        help='Display only the first 20 lines, in tabular format')

    parser.add_argument('-e', '--encoding',
                        help='Force the encoding')

    parser.add_argument('-f', '--format',
                        help="Force the file format. Typical values are 'csv', 'xls', 'xlsx' ")

    parser.add_argument('-u', '--urlfiletype',
                        help="Force the type of the file downloaded from the url. Equivalent to changing the file extension ")

    parser.add_argument('-s', '--start',
                        help='Line number where data starts')

    parser.add_argument('-d', '--headers', default=None, action='store_true',
                        help="Comma seperated list of header line numebrs")

    parser.add_argument('-E', '--enumerate', default=None, action='store_true',
                        help="Download the URL and enumerate it's contents as URLs")

    parser.add_argument('-i', '--intuit', default=None, action='store_true',
                        help="Intuit headers, start lines, etc")

    parser.add_argument('-I', '--info', default=None, action='store_true',
                        help="Print information about the url")

    parser.add_argument('-T', '--table', action='store_true',
                        help="When generating rows, output 20 rows in a table form. Otherwise, output CSV")

    parser.add_argument('url')

    cache = get_cache()

    args = parser.parse_args(sys.argv[1:])

    if args.info:
        from rowgenerators import parse_app_url
        u = parse_app_url(args.url)
        d = u.dict
        for p in u._app_parts:
            v = getattr(u,p)
            if v is not None:
                d[p] = v

        rows = list(d.items())
        rows.append(('start', u.start))
        rows.append(('headers', u.headers))
        prt(tabulate(rows))
        sys.exit(0)

    ss = Url(url=args.url, target_format=args.format, encoding=args.encoding, resource_format=args.urlfiletype)


    if args.enumerate:
        contents = list(enumerate_contents(ss, cache_fs=cache))

        for s in contents:
            print(s.rebuild_url())

    elif args.intuit:
        contents = list(enumerate_contents(ss, cache_fs=cache))
        for s in contents:

            try:
                encoding, ri = run_row_intuit(s.rebuild_url(), cache=cache)

                prt("{} headers={} start={} encoding={}".format(
                    s.rebuild_url(),
                    ','.join(str(e) for e in ri.header_lines),
                    ri.start_line,
                    encoding))
            except SourceError as e:
                warn("{}: {}".format(s.rebuild_url(), e))

    else:
        from rowgenerators import parse_app_url

        kwargs = {}
        if args.format:
            kwargs['target_format'] = args.format

        if args.encoding:
            kwargs['encoding'] = args.encoding

        if args.urlfiletype:
            kwargs['resource_format'] = args.urlfiletype

        t = parse_app_url(args.url, **kwargs).get_resource().get_target()

        rg = t.generator

        if args.table:
            print(tabulate(islice(rg, 20)))
        else:

            import csv

            w = csv.writer(sys.stdout)

            w.writerows(rg)

def listrowgen():
    import argparse

    parser = argparse.ArgumentParser(
        prog='rowgen-generators',
        description='List registered generators')

    args = parser.parse_args(sys.argv[1:])

    entries = []
    for ep in iter_entry_points('rowgenerators'):
        try:
            c = ep.load()
        except Exception as e:
            warn('Error:', ep.name, e)
        entries.append([ep.name, ep.module_name, c.__name__, ])


    print(tabulate(sorted(entries), ['EP Name', 'Module', 'Class']))



# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT, included in this distribution as LICENSE

""" """

from .file import FileUrl


class CsvFileUrl(FileUrl):
    """URL that references a CSV file"""

    match_priority = FileUrl.match_priority - 5

    @classmethod
    def _match(cls, url, **kwargs):
        return url.proto == 'file' and url.target_format == 'csv'


    @property
    def resource_url(self):

        from rowgenerators.appurl.util import unparse_url_dict

        return unparse_url_dict(self.__dict__,
                                scheme=self.scheme if self.scheme else 'file',
                                scheme_extension=False,
                                fragment=False)

    @property
    def target_format(self):
        return 'csv'




"""


Copyright (c) 2015 Civic Knowledge. This file is licensed under the terms of
the Revised BSD License, included in this distribution as LICENSE.txt

"""

import geoid
import geoid.acs
import geoid.census
import geoid.civick
import geoid.tiger

from rowgenerators.valuetype import (FailedValue, GeoMixin, IntDimension, FloatDimension, LabelValue, ROLE,
                                     NoneValue, ValueType, LOM)
from rowgenerators.valuetype import StrDimension


class FailedGeoid(FailedValue):
    def __str__(self):
        return 'invalid'


class Geoid(StrDimension, GeoMixin):
    """General Geoid """
    desc = 'General Geoid'
    geoid_cls = None # Set in derived classes
    geoid = None

    def __new__(cls, *args, **kwargs):
        import geoid

        v = args[0]

        if v is None or (isinstance(v, str) and v.strip() == ''):
            return NoneValue

        try:
            if isinstance(v, geoid.core.Geoid):
                _geoid = v
                _geoid_str = str(v)
            elif len(args) < 2:  # Parse a string
                _geoid = cls.geoid_cls.parse(v)
                _geoid_str = v
            else:  # construct from individual state, county, etc, values
                _geoid = cls.geoid_cls(*args, **kwargs)
                _geoid_str = v
        except ValueError as e:
            return FailedValue(args[0], e)

        o = super(Geoid, cls).__new__(cls, _geoid_str)
        o.geoid = _geoid

        return o


    def __getattr__(self, item):
        """Allow getting attributes from the internal geoid"""
        try:
            return getattr(self.geoid, item)
        except AttributeError:
            return object.__getattribute__(self, item)

    @property
    def acs(self):
        return self.geoid.convert(geoid.acs.AcsGeoid)

    @property
    def gvid(self):
        return self.geoid.convert(geoid.civick.GVid)

    @property
    def census(self):
        return self.geoid.convert(geoid.census.CensusGeoid)

    @property
    def tiger(self):
        return self.geoid.convert(geoid.tiger.TigerGeoid)


class GeoLabel(LabelValue, GeoMixin):
    role = ROLE.LABEL
    vt_code = 'geo/label'
    desc = 'Geographic Identifier Label'


class GeoAcs(Geoid):
    role = ROLE.DIMENSION
    vt_code = 'geoid'
    desc = 'ACS Geoid'
    geoid_cls = geoid.acs.AcsGeoid


class GeoidAcsCounty(GeoAcs):
    """An ACS Geoid for Counties """
    desc = 'County ACS geoid'
    geoid_cls = geoid.acs.County


class GeoidAcsTract(GeoAcs):
    """An ACS Geoid for Counties """
    vt_code = 'geo/acs/tract'
    desc = 'Tract ACS geoid'
    geoid_cls = geoid.acs.Tract


class GeoidCensusTract(Geoid):
    """A Census Geoid for Counties """
    desc = 'Census Tract geoid'
    vt_code =  "geoid/census/tract"
    geoid_cls = geoid.census.Tract

    @property
    def dotted(self):
        """Return just the tract number, excluding the state and county, in the dotted format"""
        v = str(self.geoid.tract).zfill(6)
        return v[0:4] + '.' + v[4:]


def county(state, county):
    return GeoidAcsCounty(state, county)


class GeoidTiger(Geoid):
    role = ROLE.DIMENSION
    vt_code = 'geo/tiger'
    desc = 'Tigerline format Geoid'
    geoid_cls = geoid.tiger.TigerGeoid

class GeoidTigerTract(Geoid):
    role = ROLE.DIMENSION
    vt_code = 'geo/tiger/tract'
    desc = 'Tigerline format GeoidAcsTract'
    geoid_cls = geoid.tiger.Tract


class GeoidCensus(Geoid):
    role = ROLE.DIMENSION
    vt_code = 'geoid/census/census'
    desc = 'Census Geoid'
    geoid_cls = geoid.census.CensusGeoid

    @classmethod
    def subclass(cls, vt_code, vt_args):
        """Return a dynamic subclass that has the extra parameters built in"""
        from geoid.geoid.core import get_class
        import geoid.census

        parser = get_class(geoid.census, vt_args.strip('/')).parse

        cls = type(vt_code.replace('/', '_'), (cls,), {'vt_code': vt_code, 'parser': parser})
        globals()[cls.__name__] = cls
        assert cls.parser

        return cls


class GeoidGvid(Geoid):
    role = ROLE.DIMENSION
    vt_code = 'geo/gvid'
    desc = 'CK Geoid'
    geoid_cls = geoid.civick.GVid


class ZipCode(IntDimension, GeoMixin):
    """A ZIP code"""

    desc = 'ZIP Code'
    vt_code = 'zip'


class ZipCodePlusFour(StrDimension, GeoMixin):
    """A ZIP code"""

    desc = 'ZIP Code with 4 digit extension'
    vt_code = 'zipp4'

class Stusab(StrDimension, GeoMixin):
    """A 2 character state abbreviation"""
    desc = 'USPS State Code'
    vt_code = 'geo/usps/state'

    def __new__(cls, v):

        if v is None:
            return NoneValue

        try:
            return str.__new__(cls, str(v).lower())
        except Exception as e:
            return FailedValue(v, e)


class Fips(IntDimension, GeoMixin):
    """A FIPS Code"""
    role = ROLE.DIMENSION
    desc = 'Fips Code'
    vt_code = 'fips'


class FipsState(IntDimension, GeoMixin):
    """A FIPS Code"""
    role = ROLE.DIMENSION
    desc = 'Fips State Code'
    vt_code = 'fips/state'

    @property
    def geoid(self):
        import geoid.census
        v = geoid.census.State(int(self))
        if not v:
            return NoneValue
        else:
            return v

class GeoInt(IntDimension, GeoMixin):
    """General integer Geo identifier"""
    role = ROLE.DIMENSION
    desc = 'General integer Geo identifier'
    vt_code = 'geo/int'


class GnisValue(IntDimension, GeoMixin):
    """An ANSI geographic code"""
    role = ROLE.DIMENSION
    desc = 'US Geographic Names Information System  Code'
    vt_code = 'gnis'


class CensusValue(IntDimension, GeoMixin):
    """An geographic code defined by the census"""
    role = ROLE.DIMENSION
    desc = 'Census Geographic Code'
    vt_code = 'geo/census'


class WellKnownTextValue(StrDimension, GeoMixin):
    """Geographic shape in Well Known Text format"""
    role = ROLE.DIMENSION
    desc = 'Well Known Text'
    vt_code = 'wkt'


# Not sure how to make this a general object, so it is a
# single element tuple
class ShapeValue(tuple, ValueType):
    _pythontype = object
    desc = 'Shape object'
    vt_code = 'geometry'
    lom = LOM.NOMINAL

    def __new__(cls, v):

        if v is None or v is NoneValue or v == '':
            return NoneValue

        try:
            return tuple.__new__(cls, [v])
        except Exception as e:
            return FailedValue(v, e)

    @property
    def shape(self):
        from shapely.wkt import loads
        from shapely.geometry.base import BaseGeometry
        if isinstance(self[0], BaseGeometry):
            return self[0]
        else:
            return loads(self[0])

    def __str__(self):
        return str(self.shape)


class DecimalDegreesValue(FloatDimension, GeoMixin):
    """An geographic code defined by the census"""
    role = ROLE.DIMENSION
    desc = 'Geographic coordinate in decimal degrees'



geo_value_types = {
    "label/geo": GeoLabel,
    "geoid": GeoAcs,  # acs_geoid
    "geoid/census": GeoAcs,  # acs_geoid
    GeoidTigerTract.vt_code: GeoidTigerTract,
    GeoidCensusTract.vt_code: GeoidCensusTract,
    "geoid/tract": GeoidAcsTract,
    GeoidAcsTract.vt_code: GeoidAcsTract,
    "geoid/county": GeoidAcsCounty,  # acs_geoid
    GeoidAcsCounty.vt_code: GeoidAcsCounty,  # acs_geoid
    "gvid": GeoidGvid,
    "fips": Fips,
    "fips/state": FipsState,  # fips_state
    "fips/county": Fips,  # fips_
    "geo/int": GeoInt,
    "gnis": GnisValue,
    "census": CensusValue,  # Census specific int code, like FIPS and ANSI, but for tracts, blockgroups and blocks
    "zip": ZipCode,  # zip
    "zipp4": ZipCodePlusFour,  # zip
    "zcta": ZipCode,  # zip
    "stusab": Stusab,  # stusab
    "lat": DecimalDegreesValue,  # Decimal degrees
    "lon": DecimalDegreesValue,  # Decimal degrees
    "wkt": WellKnownTextValue  # WKT Geometry String
}



# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT, included in this distribution as LICENSE

""" """

from os.path import basename, join
from urllib.parse import unquote

from .util import file_ext, parse_url_to_dict, unparse_url_dict

def match_url_classes(u_str, **kwargs):
    """
    Return the classes for which the url matches an entry_point specification, sorted by priority

    :param u_str: Url string
    :param kwargs: arguments passed to Url constructor
    :return:
    """

    from pkg_resources import iter_entry_points

    u = Url(str(u_str), downloader=None, **kwargs)

    try:
        classes = []

        for ep in iter_entry_points(group='appurl.urls'):
            if u._match_entry_point(ep.name):
                classes.append(ep.load())

        classes = sorted(classes, key=lambda cls: cls.match_priority)

    except ModuleNotFoundError as e:
        raise ModuleNotFoundError("Failed to find module for url string '{}', entrypoint: "
                                  .format(u_str, e))

    return classes

default_downloader = None

def parse_app_url(u_str, downloader='default', **kwargs):
    """
    Parse a URL string and return a Url object, with the class based on the highest priority
    entry point that matches the Url and which of the entry point classes pass the match() test.

    :param u_str: Url string
    :param downloader: Downloader object to use for downloading objects.
    :param kwargs: Args passed to the Url constructor.
    :return:
    """
    from rowgenerators.appurl.web.download import Downloader
    from rowgenerators.exceptions import AppUrlError

    if not u_str:
        return None

    if isinstance(u_str, Url):
        return u_str

    if not isinstance(u_str, str):
        raise AppUrlError("Input isn't a string nor Url")

    if downloader == 'default':
        global default_downloader
        if default_downloader is None:
            default_downloader = Downloader.get_instance()

        downloader = default_downloader

    classes = match_url_classes(u_str, **kwargs)

    u = Url(str(u_str), downloader=None, **kwargs)

    for cls in classes:
        if cls._match(u):
            return cls(str(u_str) if u_str else None, downloader=downloader, **kwargs)


class UrlPartsProp(object):
    """Property descriptor for reading and writting to the _parts dict
    in UrlParts"""
    def __init__(self, name):
        self.name = name

    def __get__(self, obj, objtype):
        return obj._parts.get(self.name)

    def __set__(self, obj, value):
        if value is None and self.name in obj._parts:
            del obj._parts[self.name]
        else:
            obj._parts[self.name] = value

    def __delete__(self, obj):
       del obj._parts[self.name]

class UrlParts(object):
    """Container class for handling property accessors"""

    _url_parts = ['proto', 'scheme_extension', 'scheme',
                 'netloc', 'hostname',
                 'username', 'password', 'port',
                 'path', 'query', 'fragment', 'fragment_query']

    _app_parts = ['resource_file', 'resource_format',
                 'target_file', 'target_format', 'target_segment']

    _fragment_query_parts = ['start','end','headers','encoding',
                             'resource_file','resource_format','target_format']

    _fragment_segments_parts = ['target_file','target_segment']

    _all_parts = set(_url_parts+_app_parts + _fragment_query_parts + _fragment_segments_parts )

    # Add extra fragment parts here.
    _extra_fragement_props = []

    def __init__(self, url, **kwargs):

        self._url = url
        self._kwargs = kwargs

        if self._url:
            self._parts = parse_url_to_dict(self._url)
        else:
            self._parts = {}

        if  kwargs.get('no_netloc') is True:
            self.remove_netloc()

        if kwargs.get('no_abs_path') is True:
            self.relify_path()

        self._convert_fragment()
        self._convert_fragment_query()

        self._parts.update(kwargs)

    def remove_netloc(self):
        '''If the URL has a netloc, move it to the path'''

        if self._parts.get('netloc'):
            path = self._parts['path']
            if path.startswith('/'):
                path = path[1:]
            self._parts['path'] = join(self._parts['netloc'], path)
            self._parts['netloc'] = ''

        return self

    def relify_path(self):
        """If the path is absolute, make it relative, by removing a leading /"""

        if self._parts['path'].startswith('/'):
            self._parts['path'] = self._parts['path'][1:]

        return self

    def absify_path(self):
        """If the path is relative, make it relative, by adding a leading /"""

        if not self._parts['path'].startswith('/'):
            self._parts['path'] = '/'+self._parts['path']

        return self

    def _convert_fragment(self):

        if 'fragment' in self._parts and isinstance(self._parts['fragment'], (list, tuple)):
            if len(self._parts['fragment']) == 1:
                self._parts['target_file'] = self._parts['fragment'][0]
            elif len(self._parts['fragment']) == 2:
                self._parts['target_file'], self._parts['target_segment'] = self._parts['fragment']

            del self._parts['fragment']

    def _convert_fragment_query(self):

        if isinstance(self._parts.get('fragment_query'), dict):

            for k, v in list(self._parts['fragment_query'].items()):
                if k in self._fragment_query_parts:
                    self._parts[k] = self._parts['fragment_query'][k]
                    del self._parts['fragment_query'][k]

    scheme = UrlPartsProp('scheme')
    scheme_extension = UrlPartsProp('scheme_extension')
    netloc = UrlPartsProp('netloc')
    hostname = UrlPartsProp('hostname')
    username = UrlPartsProp('username')
    password = UrlPartsProp('password')
    port = UrlPartsProp('port')
    path = UrlPartsProp('path')
    query = UrlPartsProp('query')
    target_segment = UrlPartsProp('target_segment')
    start = UrlPartsProp('start')
    end = UrlPartsProp('end')
    headers = UrlPartsProp('headers')
    encoding = UrlPartsProp('encoding')

    fragment_query = UrlPartsProp('fragment_query')

    @property
    def proto(self):
        return self._parts.get('proto') or \
               self._parts['scheme_extension'] or \
               {'https': 'http', '': 'file'}.get(self._parts['scheme']) or \
               self._parts['scheme']

    @proto.setter
    def proto(self,v):
        self._parts['proto'] = v

    @property
    def target_format(self):
        from .util import file_ext

        target_format = self._parts.get('target_format')

        if not target_format and self.target_file:
            target_format = file_ext(self.target_file)

        if not target_format:
            target_format = self.resource_format

        # handle URLS that end with package names, like:
        # 'example.com-example_data_package-2017-us-1'
        if target_format and len(target_format) > 8:
            target_format = None

        return target_format

    @target_format.setter
    def target_format(self, v):
        self._parts['target_format'] = v



    def clear_fragment(self):
        """
        Return a copy of the URL with no fragment components

        :return: A cloned URl object, with the fragment and fragment queries cleared.
        """

        c = self.clone()
        c._parts['target_file'] = None
        c._parts['target_segment'] = None

        return c

    #
    # Property accessors
    #

    def set_fragment(self, f):
        """Return a clone with the fragment set"""
        raise NotImplementedError()

    @property
    def resource_file(self):
        if self.path:
            return basename(self.path)
        else:
            return None

    @property
    def resource_format(self):
        return self._parts.get('resource_format') or file_ext(self.resource_file)

    @resource_format.setter
    def resource_format(self, v):
        self._parts['resource_format'] = v

    @property
    def target_file(self):

        return self._parts.get('target_file') or self.resource_file

    @target_file.setter
    def target_file(self, v):
        self._parts['target_file'] = v


    def set_target_file(self, v):
        """Return a clone with a target_file set"""
        u = self.clone()
        u.target_file = v
        return u

    def set_target_segment(self, v):
        """Return a clone with a target_file set"""
        u = self.clone()
        u.target_segment = v
        return u

    @property
    def dict(self):
        """
        Returns a dictionary of the object components.

        :return: a dict.
        """

        d = dict(self._parts.items())

        d['scheme_extension'] = self._parts.get('proto') or d.get('scheme_extension')

        for k, v in list(d.items()):
            if k in (self._fragment_query_parts + self._fragment_segments_parts):
                if not v:
                    del d[k]

        d['fragment'] = [
            self._parts.get('target_file'),
            self._parts.get('target_segment')
        ]

        for k in self._fragment_query_parts:
            if k in d:
                d['fragment_query'][k] = d[k]
                del d[k]

        return d

    @property
    def frag_dict(self):
        d = {}
        for k in self._fragment_segments_parts + self._fragment_query_parts:
            d[k] = self._parts.get(k)

        return d

    def __str__(self):

        return unparse_url_dict(self.dict)




class Url(UrlParts):
    """Base class for Application URLs .

    After construction, a Url object has a set of properties and attributes for access
    the parts of the URL, and method for manipulating it. The attributes and properties
    include the typical properties of a parsed URL, plus properties that are derives from the
    typical parts, and a few extra components that can be part of the fragment query.

    The typical parts are:

    - ``scheme``
    - ``scheme_extension``
    - ``netloc``
    - ``hostname``
    - ``path``
    - ``params``
    - ``query``
    - ``fragment``
    - ``username``
    - ``password``
    - ``port``

    The ``fragment`` is special; it is an array of two elements, the first of which is the ``target_file`` and
    and the second is the ``target_segment``. If there are other parts of the source URL, they must be
    formates as queriy components, and will be parsed into the ``fragment_query``.

    Special application components are:

    - ``proto``. This is set to the ``scheme_extension`` if it exists, the scheme otherwise.
    - ``resource_file``. The filename of the resource to download. It is usually the last part of the URL, but can be overidden in the fragment
    - ``resource_format``. The format name of the resource, normally drawn from the ``resoruce_file`` extension, but can be overidden in the fragment
    - ``target_file``. The filename of the file that will be produced by :py:meth`Url.get_target`, but may be overidden.
    - ``target_format``. The format of the ``target_file``, but may be overidden.
    - ``target_segment``. A sub-component of the ```target_file``, such as the worksheet in a spreadsheet.
    - ``fragment_query``. Holds additional parts of the fragment.

    When the fragment holds extra parts, these can be be formatted as a URL query. Recognized keys are:

    - ``resource_file``
    - ``resource_format``
    - ``target_file``
    - ``target_format``
    - ``encoding``. Text encoding to be used when reading the target.
    - ``headers``. For row-oriented data, the row numbers of the headers, as a comma-seperated list of integers.
    - ``start``. For row-oriented data, the row number of the first row of data ( as opposed to headers. )
    - ``end``. For row-oriented data, the row number of the last row of data.

    """

    match_priority = 100
    match_proto = None
    generator_class = None  # If set, generators match with name = <{generator_class}>

    def __init__(self, url=None, downloader=None, no_netloc=False, no_abs_path=False,
                 **kwargs):
        """  Initialize a new Application Url

        :param url: URL string
        :param downloader: :py:class:`appurl.web.download.Downloader` object.
        :param no_netloc: If true, assume there is no netloc component, so 'proto://' and 'proto:/' are
            equivalent Will move any netloc component into the path
        :param no_abs_path: If true, convert all paths to be relative -- remove any leading '/'
        :param kwargs: Additional arguments override URL properties.
        :return: An Application Url object

        Keyword arguments will override properties set by parsing the URL string.

        """

        self._kwargs = kwargs
        self._downloader = downloader

        super().__init__(url, no_netloc=no_netloc, no_abs_path=no_abs_path, **kwargs)

        assert 'is_archive' not in self._kwargs #?

    def resolve(self):
        """Resolve a URL to another format, such as by looking up a URL that specified a
        search, into another URL. The default implementation returns self. """
        return self

    def get_resource(self):
        """Get the contents of resource and save it to the cache, returning a file-like object"""
        raise NotImplementedError(("get_resource not implemented in {} for '{}'. "
                                   "You may need to install a python mpdule for this type of url")
                                  .format(self.__class__.__name__, str(self)))

    def get_target(self):
        """Get the contents of the target, and save it to the cache, returning a file-like object
        """
        raise NotImplementedError(("get_target not implemented in {} for '{}'"
                                   "You may need to install a python module for this type of url"
                                   )
                                  .format(self.__class__.__name__, str(self)))

    @property
    def downloader(self):
        """Return the Downloader() for this URL"""
        return self._downloader

    def list(self):
        """Return URLS for files contained in an container. This implementation just returns
        ``[self]``, but sub classes may, for instance, list all of the sub-components of a directory,
        or all of the worksheets in an Excel file. """
        return [self]

    @property
    def is_archive(self):
        """Return true if this URL is for an archive. Currently only ZIP is recognized"""
        return self.resource_format in self.archive_formats

    # property
    def archive_file(self):
        """Return the name of the archive file, if there is one."""
        return self.target_file if self.is_archive and self.resource_file != self.target_file else None

    @property
    def fspath(self):
        """The path in a form suitable for use in a filesystem"""
        from pathlib import PurePath
        return PurePath(unquote(self.path))

    @property
    def path_is_absolute(self):
        return self.path.startswith('/')

    def join(self, s):
        """ Join a component to the end of the path, using :func:`os.path.join`. The argument
        ``s`` may be a :class:`appurl.Url` or a string. If ``s`` includes a ``netloc`` property,
        it is assumed to be an absolute url, and it is returned after parsing as a Url. Otherwise,
        the path component of ``s`` is extracted and joined to the path component of this url.

        :param s: A Url object, or a string.
        :return: A copy of this url.
        """

        from copy import copy
        import pathlib

        try:
            path = s.path
            netloc = s.netloc
            u = s
        except AttributeError:
            u = parse_app_url(s, downloader=self.downloader)
            path = u.path
            netloc = u.netloc

        # If there is a netloc, it's an absolute URL
        if netloc:
            return u

        url = copy(self)

        # Using pathlib.PurePosixPath ensures using '/' on windows. os.path.join will use '\'
        url.path = str(pathlib.PurePosixPath(self.path).joinpath(path))

        return url

    def join_dir(self, s):
        """ Join a component to the parent directory of the path, using join(dirname())

        :param s:
        :return: a copy of this url.
        """

        from os.path import dirname
        from copy import copy
        import pathlib

        try:
            path = s.path
            netloc = s.netloc
            u = s
        except AttributeError:
            u = parse_app_url(s, downloader=self.downloader)
            path = u.path
            netloc = u.netloc

        # If there is a netloc, it's an absolute URL
        if netloc:
            return u

        url = copy(self)
        # Using pathlib.PurePosixPath ensures using '/' on windows. os.path.join will use '\'
        url.path = str(pathlib.PurePosixPath(dirname(self.path)).joinpath(path))

        return url

    def join_target(self, tf):
        """Return a new URL, possibly of a new class, with a new target_file"""

        raise NotImplementedError("Not implemented in '{}' ".format(type(self)))

    @property
    def inner(self):
        """Return the URL without the scheme extension and fragment. Re-parses the URL, so it should return
        the correct class for the inner URL. """
        if not self.scheme_extension:
            return self

        c = self.clone(scheme_extension=None, proto=None)

        return parse_app_url(str(c), downloader=self.downloader)


    @property
    def resource_url(self):

        return unparse_url_dict(self.dict,
                                scheme=self.scheme if self.scheme else 'file',
                                scheme_extension=False,
                                fragment_query=False,
                                fragment=False)

    def dirname(self):
        """Return the dirname of the path"""
        from os.path import dirname

        u = self.clone()
        u.path = dirname(self.path)
        return u



    def as_type(self, cls):
        """
        Return the URL transformed to a different class. Copies the downloader and
        build the new url using :py:meth:`Url.dict`

        :param cls: Class of Url to construct
        :return: A new Url object
        """

        return cls(downloader=self.downloader, **self.dict)

    def interpolate(self, context=None):
        """
        Use the Downloader.context to interpolate format strings in the URL. Re-parses the URL,
         returning a new URL

        :param context: Extra context to interpolate with
        :return:
        """

        from copy import copy

        cxt = copy(self.downloader.context)

        cxt.update(context or {})

        from rowgenerators.exceptions import AppUrlError

        try:
            return parse_app_url(str(self).format(**cxt), downloader=self.downloader)
        except KeyError as e:
            raise AppUrlError("Failed to interpolate '{}'; context is {}. Missing key: {} "
                              .format(str(self), self.downloader.context, e))

    def clone(self, **kwargs):
        """
        Return a clone of this Url, possibly with some arguments replaced.

        :param kwargs: Keyword arguments are arguments to set in the copy, using :py:func:`setattr`
        :return: A cloned Url object.
        """
        from copy import deepcopy

        c = deepcopy(self)

        for k, v in kwargs.items():
            try:
                setattr(c, k, v)

            except AttributeError:
                raise AttributeError("Can't set attribute '{}' on '{}' ".format(k, c))

        return c

    @property
    def generator(self):
        """
        Return the generator for this URL, if the rowgenerator package is installed.

        :return: A row generator object.
        """

        from rowgenerators.core import get_generator

        r = self.get_resource()
        t = r.get_target()

        return get_generator(t.get_target(), source_url=self)


    #
    # Matching methods
    #

    def _match_entry_point(self, name):
        """Return true if this URL matches the entrypoint pattern

        Entrypoint patterns:

            'scheme:' Match the URL scheme
            'proto+' Matches the protocol / scheme_extension
            '.ext' Match the resource extension
            '#.ext' Match the target extension
        """

        import re

        if '&' in name:
            return all(self._match_entry_point(n) for n in name.split('&'))

        try:
            name = name.name  # Maybe it's an entrypoint entry, not the name
        except AttributeError:
            pass

        if name == '*':
            return True
        elif name.startswith("/") and name.endswith("/"):
            return re.search(name[1:-1], str(self))
        elif name.endswith(":"):
            return name[:-1] == self.scheme
        elif name.endswith('+'):
            return name[:-1] == self.proto
        elif name.startswith('.'):
            return name[1:] == self.resource_format
        elif name.startswith('#.'):
            return name[2:] == self.target_format
        else:
            return False

    @classmethod
    def _match(cls, url, **kwargs):
        """Return True if this handler can handle the input URL"""
        if cls.match_proto:
            return url.proto == cls.match_proto
        else:
            return True;  # raise NotImplementedError("Match is not implemented for class '{}' ".format(str(cls)))

    #
    # Other support methods
    #


    def __deepcopy__(self, memo):
        return type(self)(None, downloader=self._downloader, **self._parts)

    def __copy__(self):
        return type(self)(None, downloader=self._downloader, **self._parts)

    def _decompose_fragment(self, frag):
        """Parse the fragment component"""
        from urllib.parse import unquote_plus

        if isinstance(frag, (list, tuple)):
            assert frag[0] is None or isinstance(frag[0], str), (frag[0], type(frag[0]))
            return frag

        if not frag:
            return None, None

        frag_parts = unquote_plus(frag).split(';')

        if not frag_parts:
            file, segment = None, None
        elif len(frag_parts) == 1:
            file = frag_parts[0]
            segment = None
        elif len(frag_parts) >= 2:
            file = frag_parts[0]
            segment = frag_parts[1]

        assert file is None or isinstance(file, str), (file, type(file))

        return file, segment

    def __repr__(self):
        return "<{} {}>".format(self.__class__.__name__, str(self))





"""


Copyright (c) 2015 Civic Knowledge. This file is licensed under the terms of
the Revised BSD License, included in this distribution as LICENSE.txt

"""
from __future__ import print_function

from datetime import date, time, datetime

from decorator import decorator
from rowgenerators.util import Constant, memoize

from tableintuit.types import test_nan

ROLE = Constant()
ROLE.DIMENSION = 'd'
ROLE.MEASURE = 'm'
ROLE.LABEL = 'e'
ROLE.ERROR = 'e'
ROLE.KEY = 'k'
ROLE.IDENTIFIER = 'i'
ROLE.UNKNOWN = 'u'
ROLE.OTHER = 'u'

role_descriptions = {
    ROLE.DIMENSION: 'dimension',
    ROLE.MEASURE: 'measure',
    ROLE.ERROR: 'error',
    ROLE.KEY: 'key',
    ROLE.IDENTIFIER: 'identifier',
    ROLE.UNKNOWN: 'unknown',
    ROLE.OTHER: 'other'
}

# Levels of Measurement
LOM = Constant()
LOM.UNKNOWN = 'u'
LOM.NOMINAL = 'n'
LOM.ORDINAL = 'o'
LOM.INTERVAL = 'i'
LOM.RATIO = 'r'

@decorator
def valuetype(func, *args, **kw):
    return func(*args, **kw)


def count_errors(errors):
    from rowgenerators.rowpipe.exceptions import TooManyCastingErrors

    # Count at most 10 errors for each column
    c = sum(len(e) if len(e) < 10 else 10 for e in errors.values())

    if c > 50:
        raise TooManyCastingErrors("Too many casting errors ({})".format(c), errors=errors)


class TimeMixin(object):
    pass


class GeoMixin(object):
    pass


class DimensionMixin(object):
    pass


class LabelMixin(object):
    pass


class ValueType(object):
    role = ROLE.UNKNOWN
    low = LOM.UNKNOWN
    _pythontype = str
    desc = ''

    def raise_for_error(self):
        # Not an error, so nothing to raise
        pass

    @property
    def is_none(self):
        return False

    @classmethod
    def python_type(self):
        return self._pythontype

    @property
    def failed_value(self):
        return None

    @classmethod
    def description(cls):
        if cls.desc:
            return cls.desc
        elif cls.__doc__:
            return cls.desc
        else:
            return cls.vt

    @classmethod
    def subclass(cls, vt_code, vt_args):
        """
        Return a dynamic subclass that has the extra parameters built in
        :param vt_code: The full VT code, privided to resolve_type
        :param vt_args: The portion of the VT code to the right of the part that matched a ValueType
        :return:
        """
        return type(vt_code.replace('/', '_'), (cls,), {'vt_code': vt_code, 'vt_args': vt_args})

    @classmethod
    def is_time(cls):
        return issubclass(cls, TimeMixin)

    @classmethod
    def is_geo(cls):
        return issubclass(cls, GeoMixin)

    @classmethod
    def is_geoid(cls):
        from rowgenerators.valuetype import Geoid
        return issubclass(cls, Geoid)


    @classmethod
    def is_label(cls):
        return issubclass(cls, LabelMixin)

class _NoneValue(object):
    """Represent None as a ValueType"""

    _pythontype = None

    @classmethod
    def python_type(self):
        return self._pythontype

    def raise_for_error(self):
        pass

    @property
    def is_none(self):
        return True

    @property
    def failed_value(self):
        return None

    def __bool__(self):
        return False

    def __nonzero__(self):
        return False

    def __len__(self):
        return False

    def __getattr__(self, item):
        """All properties of NoneValue are also NoneValues"""
        return self

    def __repr__(self):
        return "NoneValue"


NoneValue = _NoneValue()


class FailedValue(str, ValueType):
    """When ValueTypes fail to convert, the __new__ returns an object of this type,
    which resolves as a string containing the value that failed """

    _pythontype = str
    exc = None

    def __new__(cls, *args, **kwargs):
        o = str.__new__(cls, args[0])
        return o

    def __init__(self, v, exc=None):
        self.exc = exc

    def raise_for_error(self):
        if self.exc:
            raise self.exc

    @property
    def is_none(self):
        return False

    @property
    def failed_value(self):
        return str(self)

    def __bool__(self):
        return False

    __nonzero__ = __bool__

    def __getattr__(self, item):
        return self


def cast_int(v, header_d, errors):
    if isinstance(v, FailedValue):

        errors[header_d].add(u"Failed to cast '{}' ({}) to int in '{}': {}".format(v, type(v), header_d, v.exc))
        count_errors(errors)
        return None

    if v != 0 and not bool(v):
        return None
    else:
        try:
            return int(v)
        except Exception as e:
            errors[header_d].add(u"Failed to cast '{}' ( {} ) to int in '{}': {}".format(v, type(v), header_d, e))
            count_errors(errors)


cast_id = cast_int


def cast_long(v, header_d, errors):
    if isinstance(v, FailedValue):
        errors[header_d].add(u"Failed to cast '{}' ({}) to long in '{}': {}".format(v, type(v), header_d, v.exc))
        count_errors(errors)
        return None

    if v != 0 and not bool(v):
        return None
    else:
        try:
            return int(v)
        except Exception as e:
            errors[header_d].add(u"Failed to cast '{}' ( {} ) to long in '{}': {}".format(v, type(v), header_d, e))
            count_errors(errors)


def cast_float(v, header_d, errors):
    if isinstance(v, FailedValue):
        errors[header_d].add(u"Failed to cast '{}' ({}) to float in '{}': {}".format(v, type(v), header_d, v.exc))
        count_errors(errors)
        return None

    if v != 0 and not bool(v):
        return None
    else:
        try:
            return float(v)
        except Exception as e:
            errors[header_d].add(u"Failed to cast '{}' ( {} )  to float in '{}': {}".format(v, type(v), header_d, e))
            count_errors(errors)


def cast_str(v, header_d, errors):

    if isinstance(v, FailedValue):
        errors[header_d].add(u"Uncleared errors on value '{}' in '{}': {}".format(v, header_d, str(v.exc)))
        count_errors(errors)

    if v != 0 and not bool(v):
        return None
    else:
        try:
            return str(v)
        except Exception as e:
            errors[header_d].add(u"Failed to cast '{}' ( {} )  to str in '{}': {}".format(v, type(v), header_d, e))
            count_errors(errors)


def cast_text(v, header_d, errors):
    if isinstance(v, FailedValue):
        errors[header_d].add(u"Failed to cast '{}' ({}) to unicode in '{}': {}".format(v, type(v), header_d, v.exc))
        count_errors(errors)

    if v != 0 and not bool(v):
        return None
    else:
        try:
            return str(v)
        except Exception as e:
            errors[header_d].add(u"Failed to cast '{}' ( {} )  to unicode in '{}': {}".format(v, type(v), header_d, e))
            count_errors(errors)


cast_unicode = cast_text


def cast_object(v, header_d, errors):

    return v



class StrValue(str, ValueType):
    _pythontype = str
    desc = 'Character String'
    vt_code = 'str'
    lom = LOM.NOMINAL

    def __new__(cls, v):

        if v is None or v is NoneValue or v == '':
            return NoneValue

        try:
            return str.__new__(cls, v)
        except Exception as e:
            return FailedValue(v, e)


class TextValue(str, ValueType):
    _pythontype = str
    desc = 'Text String'
    vt_code = 'text'
    low = LOM.NOMINAL

    def __new__(cls, v):

        if v is None or v is NoneValue or v == '':
            return NoneValue

        try:
            return str.__new__(cls, v)
        except Exception as e:
            return FailedValue(v, e)


class IntValue(int, ValueType):
    _pythontype = int
    desc = 'Integer'
    vt_code = 'int'

    def __new__(cls, v):

        try:
            # The pre-conversion to float handles cases like 1.0, which
            # would throw an exception otherwize
            return int.__new__(cls, float(v))
        except Exception as e:

            if v is None or v is NoneValue or v == '' or test_nan(v):
                return NoneValue

            return FailedValue(v, e)


class LongValue(int, ValueType):
    _pythontype = int
    desc = 'Long'
    vt_code = 'long'

    def __new__(cls, v):
        try:
            return int.__new__(cls, v)
        except Exception as e:

            if v is None or v is NoneValue or v == '' or test_nan(v):
                return NoneValue

            return FailedValue(v, e)


class FloatValue(float, ValueType):
    _pythontype = float
    desc = 'General Floating Point'
    vt_code = 'float'

    def __new__(cls, v):

        try:
            return float.__new__(cls, v)
        except Exception as e:

            if v is None or v is NoneValue or v == '' or test_nan(v):
                return NoneValue

            return FailedValue(v, e)


class DateValue(date, ValueType):
    _pythontype = date
    role = ROLE.DIMENSION
    desc = 'Date'
    vt_code = 'date'
    low = LOM.ORDINAL

    def __new__(cls, v):
        from dateutil import parser

        if v is None or (isinstance(v, str) and v.strip() == ''):
            return NoneValue

        try:
            if isinstance(v, (datetime, date)):
                d = v
            else:
                d = parser.parse(str(v))

            return super(DateValue, cls).__new__(cls, d.year, d.month, d.day)
        except TypeError:
            if str(v) == 'NaT':  # THe Pandas "Not A Time" value
                return NoneValue
            raise
        except ValueError as e:
            return FailedValue(v, e)


    def __init__(self, v):
        pass


class TimeValue(time, ValueType):
    _pythontype = time
    desc = 'time'
    low = LOM.ORDINAL

    def __new__(cls, v):
        from dateutil import parser

        if v is None or (isinstance(v, str) and v.strip() == ''):
            return NoneValue

        try:
            if isinstance(v, time):
                d = v
            elif not isinstance(v, str):
                d = parser.parse(str(v))
            else:
                d = parser.parse(v)

            return super(TimeValue, cls).__new__(cls, d.hour, d.minute, d.second)
        except TypeError:
            if str(v) == 'NaT':  # THe Pandas "Not A Time" value
                return NoneValue
            raise


class DateTimeValue(datetime, ValueType):
    _pythontype = datetime
    desc = 'Date/Time'
    vt_code = 'datetime'
    low = LOM.ORDINAL

    def __new__(cls, v):
        from dateutil import parser

        if v is None or (isinstance(v, str) and v.strip() == ''):
            return NoneValue

        try:
            if isinstance(v, datetime):
                d = v
            elif not isinstance(v, str):
                d = parser.parse(str(v))
            else:
                d = parser.parse(v)
            return super(DateTimeValue, cls).__new__(cls, d.year, d.month, d.day, d.hour, d.minute, d.second)

        except TypeError:
            if str(v) == 'NaT':  # THe Pandas "Not A Time" value
                return NoneValue
            raise


class KeyVT(IntValue):
    role = ROLE.KEY
    lom = LOM.ORDINAL
    vt_code = 'k'


class IdentifierVT(IntValue):
    role = ROLE.IDENTIFIER
    lom = LOM.ORDINAL
    vt_code = 'i'


class IntDimension(IntValue, DimensionMixin):
    role = ROLE.DIMENSION
    lom = LOM.ORDINAL
    vt_code = 'dimension/int'


class FloatDimension(FloatValue, DimensionMixin):
    role = ROLE.DIMENSION
    lom = LOM.ORDINAL
    vt_code = 'dimension/float'


class TextDimension(TextValue, DimensionMixin):
    role = ROLE.DIMENSION
    lom = LOM.NOMINAL
    vt_code = 'dimension/text'


class StrDimension(StrValue, DimensionMixin):
    role = ROLE.DIMENSION
    lom = LOM.NOMINAL
    vt_code = 'dimension'


class LabelValue(TextDimension, LabelMixin):
    desc = 'Value Label'
    vt_code = 'label'
    role = ROLE.DIMENSION


def null():
    return None

def empty_str():
    return ""

def upper(v):
    if isinstance(v, (FailedValue, _NoneValue)):
        return v

    return v.upper()


def lower(v):
    if isinstance(v, (FailedValue, _NoneValue)):
        return v

    return v.lower()


def title(v):
    if isinstance(v, (FailedValue, _NoneValue)):
        return v

    return v.title()

def strip(v):
    if isinstance(v, (FailedValue, _NoneValue)):
        return v

    return v.strip()


def robust_int(v):
    """Parse an int robustly, ignoring commas and other cruft. """

    if isinstance(v, int):
        return v

    if isinstance(v, float):
        return int(v)

    v = str(v).replace(',', '')

    if not v:
        return None

    try:
        return int(v)
    except ValueError:
        try:
            return int(float(v))
        except ValueError:
            return None

def robust_float(v):
    """Parse float robustly, ignoring commas and other cruft. """

    if isinstance(v, int):
        return float(v)

    if isinstance(v, float):
        return v

    v = str(v).replace(',', '')

    if not v:
        return None

    try:
        return float(v)
    except ValueError:
        return None


def print_value(row_n, header_d, v):

    print("print_value {}: {} = {}".format(row_n, header_d, v))
    return v

def print_type(row_n, header_d, v):

    print("print_type {}: {} = {}".format(row_n, header_d,type(v)))
    return v



# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT, included in this distribution as LICENSE

"""Download resoruces and generate collections of URLs with targets and segments for
the files containes in Excel and ZIP files"""



def inspect(url, callback=None):
    """Return a list of possible extensions to the url, such as files within a ZIP archive, or
    worksheets in a spreadsheet"""
    from zipfile import ZipFile
    from copy import deepcopy

    raise NotImplementedError()

    if callback:
        callback("Inspecting: format={} file={} segment={} url={}".format(
            url.target_format, url.target_file, url.target_segment, url.rebuild_url()))

    if url.is_archive and url.target_file is None:

        if url.target_file is None:
            l = []

            for file_name in real_files_in_zf(zf):
                l.append(ss.update(target_file=file_name))

            return l

        elif ss.target_format in ('xls', 'xlsx') and ss.target_segment is None:
            src = get_generator(ss, cache_fs)
            l = []
            for seg in src.children:
                ss2 = ss.update(target_segment=seg)
                l.append(ss2)

            return l

    if ss.target_format in ('xls', 'xlsx') and ss.target_segment is None:

        src = get_generator(ss, cache_fs)

        l = []
        for seg in src.children:
            ss2 = ss.update(target_segment=seg)
            l.append(ss2)

        return l

    return [deepcopy(ss)]

def enumerate_contents(base_spec, cache_fs, callback=None):
    """Inspect the URL, and if it is a container ( ZIP Or Excel ) inspect each of the contained
    files. Yields all of the lower-level URLs"""

    from rowgenerators.appurl.url import Url, parse_app_url

    if not isinstance(base_spec, Url):
        base_spec = parse_app_url(url=base_spec)

    for s in inspect(base_spec, cache_fs, callback=callback):
        for s2 in inspect(s, cache_fs, callback=callback):
            yield s2




# Copyright (c) 2016 Civic Knowledge. This file is licensed under the terms of the
# MIT License, included in this distribution as LICENSE.txt

"""

Row generating row processor

"""

from collections import defaultdict
from rowgenerators.rowpipe.codegen import make_row_processors, exec_context
from .exceptions import RowProcessorError


class RowProcessor(object):
    """
    """

    def __init__(self, source, dest_table, source_headers=None, env=None, manager=None, code_path=None):

        """

        :param source: Row generating soruce
        :param dest_table: Destination table
        :param source_headers:
        :param env:
        :param env: A higher-level controller object, to be referenced from user-written transforms.
        :return:
        """

        if not source:
            raise RowProcessorError("No source")

        if not dest_table:
            raise RowProcessorError("No destination row processor table")

        self.source = source
        self.manager = manager
        self.source_headers = source_headers if source_headers is not None else self.source.headers
        self.dest_table = dest_table
        self.code_path = code_path

        self.env = exec_context()

        if env is not None:
            self.env.update(env)

        self.env['source'] = self.source
        self.env['pipe'] = None

        self.scratch = {}
        self.accumulator = {}
        self.errors = defaultdict(set)

        self.code = make_row_processors(self.source_headers, self.dest_table, env=self.env)

        self.code_path = self.write_code()

        exec (compile(self.code, self.code_path, 'exec'), self.env)

        self.procs = self.env['row_processors']

    def write_code(self):
        import hashlib
        import os
        import tempfile

        if self.code_path:
            path = self.code_path

        else:
            tf = tempfile.NamedTemporaryFile(prefix="rowprocessor-",
                                             suffix='{}.py'.format(hashlib.md5(self.code.encode('utf-8')).hexdigest()),
                                             delete=False)
            path = tf.name
            tf.close()

        if not os.path.exists(os.path.dirname(path)):
            os.makedirs(os.path.dirname(path))

        with open(path,'w') as f:
            f.write(self.code)

        return path

    @property
    def headers(self):
        """Return a list of the names of the columns of this file, or None if the header is not defined.

        This should *only* return headers if the headers are unambiguous, such as for database tables,
        or shapefiles. For other files, like CSV and Excel, the header row can not be determined without analysis
        or specification."""

        return self.dest_table.headers

    @headers.setter
    def headers(self, v):
        raise NotImplementedError()

    @property
    def meta(self):
        return {}

    def __iter__(self):
        """Iterate over all of the lines in the file"""
        from rowgenerators.rowproxy import RowProxy

        self.start()

        pipe = self.env['pipe']

        rp1 = RowProxy(self.source_headers) # The first processor step uses the source row structure
        rp2 = RowProxy(self.dest_table.headers) # Subsequent steps use the dest table

        for i, row in enumerate(self.source):

            try:
                rp = rp1

                for proc in self.procs:
                    row = proc(rp.set_row(row), i, self.errors, self.scratch, self.accumulator,
                               pipe, self.manager, self.source)

                    # After the first round, the row has the destination headers.
                    rp = rp2

                yield row
            except Exception as e:
                raise RowProcessorError("Exception at source ({}) row {}: {}".format(type(self.source), i, str(e))) from e


        self.finish()

    def start(self):
        pass

    def finish(self):
        pass


# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT, included in this distribution as LICENSE

"""
File URLs represent resources that are acessible on the local file system.

Because these URLs are assumed to be local, the resource already exists on
the local file system, and the :py:meth:`FileUrl.get_resource()` just returns ``self``

"""


from .file import FileUrl
from .program import ProgramUrl
from .python import PythonUrl
from .csv import CsvFileUrl
from .excel import ExcelFileUrl
from .hdf5 import Hdf5Url
#from .shapefile import ShapefileShpUrl, ShapefileUrl


# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT License, included in this distribution as LICENSE.txt

""" """

import sys
import os
from rowgenerators.source import Source
import json

class JsonRowSource(Source):
    """Generate rows from a JSON file with an array of rows. """

    delimiter = ','

    def __init__(self, ref, cache=None, working_dir=None, env=None, **kwargs):
        super().__init__(ref, cache, working_dir, **kwargs)

        self.url = ref

        if not self.url.exists():
            raise FileNotFoundError(self.url)

        if self.url.scheme != 'file':
            assert self.url.scheme == 'file', str(self.url)

        self._meta = {}

    def __iter__(self):
        """Iterate over all of the lines in the file"""


        self.start()

        try:

            # Python 3.6 considers None to mean 'utf8', but Python 3.5 considers it to be 'ascii'
            encoding = self.url.encoding or 'utf8'

            with open(self.url.fspath, encoding=encoding) as f:
                yield from json.load(f)

        except UnicodeError as e:
            raise

        self.finish()

    def finish(self):
        super().finish()


    @property
    def headers(self):
        '''Return the headers. This implementation just returns the first line, which is not always correct'''

        return next(iter(self))


    def dataframe(self, limit=None, *args, **kwargs):
        import pandas

        i = iter(self)

        columns =  next(i)

        rows = list(i)

        return pandas.DataFrame(rows, columns=columns)



# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT License, included in this distribution as LICENSE.txt

""" """

from itertools import islice

from rowgenerators.util import md5_file

from .appurl.web.download import Downloader


class RowGenerator(object):
    """Main class for accessing row generators"""

    def __init__(self, url, *args, downloader=None, **kwargs):
        from .appurl.url import parse_app_url

        self._url_text = url
        self._downloader = downloader or Downloader.get_instance()

        self.url = parse_app_url(self._url_text, *args, downloader=self._downloader, **kwargs)

    def registered_urls(self):
        """Return an array of registered Urls. The first row is the header"""
        from pkg_resources import iter_entry_points

        entries = ['Priority', 'EP Name', 'Module', 'Class']
        for ep in iter_entry_points('appurl.urls'):
            c = ep.load()
            entries.append([c.match_priority, ep.name, ep.module_name, c.__name__, ])

        return entries

    def __iter__(self):
        """Yields first the header, then each of the data rows. """
        yield from self.url.generator

    @property
    def iter_dict(self):
        """Iterate over dicts"""
        yield from self.url.generator.iter_dict

    @property
    def iter_row(self):
        """Iterate, yielding row proxy objects. DOes not first yield a header"""
        yield from self.url.generator.iter_rp

    @property
    def generator(self):
        """Return the data generating object"""
        return self.url

    @property
    def headers(self):
        """Return the columns headers"""
        return self.generator.headers

    def dataframe(self, *args, **kwargs):
        """Return a pandas dataframe"""

        try:
            return self.url.generator.dataframe(*args, **kwargs)
        except AttributeError:
            pass

        try:
            return self.url.dataframe(*args, **kwargs)
        except AttributeError:
            pass

        raise NotImplementedError("Url '{}' of type '{}' can't generate a dataframe ".format(self.url, type(self.url)))

    def geoframe(self, *args, **kwargs):
        """Return a Geopandas dataframe"""

        try:
            return self.url.geoframe(*args, **kwargs)
        except AttributeError:
            pass

        try:
            return self.url.geo_generator.geoframe(*args, **kwargs)
        except AttributeError:
            pass

        try:
            return self.url.generator.geoframe(*args, **kwargs)
        except AttributeError:
            pass

        raise NotImplementedError("Url '{}' of type '{}' can't generate a dataframe ".format(self.url, type(self.url)))

    def intuit(self):
        """Return information about the columns, based on guessing data types"""
        raise NotImplemented()

    def statistics(self):
        """Return summary statistics for the columns"""
        raise NotImplemented()

    def set_row_processor(self):
        """Register a row processor, which will transform rows as they are iterated"""
        raise NotImplemented()


class Source(object):
    """Base class for accessors that generate rows from any source. This is the class returned from
     parse_app_url().generator

    Subclasses of Source must override at least _get_row_gen method.
    """

    priority = 100

    def __init__(self, ref, cache=None, working_dir=None, env=None, **kwargs):
        self.ref = ref

        self.cache = cache

        self._meta = {}

    @property
    def headers(self):
        """Return a list of the names of the columns of this file, or None if the header is not defined.

        This should *only* return headers if the headers are unambiguous, such as for database tables,
        or shapefiles. For other files, like CSV and Excel, the header row can not be determined without analysis
        or specification."""

        return None

    @headers.setter
    def headers(self, v):
        """Catch attempts to set"""
        raise NotImplementedError

    @property
    def columns(self):
        """ Returns columns for the file accessed by accessor.

        """

        return None

    @property
    def meta(self):
        return self._meta

    @property
    def hash(self):
        with open(self.url.fspath, 'rb') as f:
            return md5_file(f)

    def __iter__(self):
        """Iterate over all of the lines in the file"""

        raise NotImplementedError()

    @property
    def iter_rp(self):
        """Iterate, yielding row proxy objects rather than rows"""

        from .rowproxy import RowProxy

        itr = iter(self)

        headers = next(itr)

        row_proxy = RowProxy(headers)

        for row in itr:
            yield row_proxy.set_row(row)

    @property
    def iter_dict(self):
        """Iterate, yielding dicts rather than rows"""

        itr = iter(self)

        headers = next(itr)

        for row in itr:
            yield dict(zip(headers, row))

    def dataframe(self, *args, **kwargs):
        """Return a pandas dataframe from the resource"""

        from pandas import DataFrame

        # Just normal data, so use the iterator in this object.
        headers = next(islice(self, 0, 1))
        data = islice(self, 1, None)

        return DataFrame(list(data), columns=headers)

    def start(self):
        pass

    def finish(self):
        pass


class SelectiveRowGenerator(object):
    """Proxies an iterator to remove headers, comments, blank lines from the row stream.
    The header will be emitted first, and comments are available from properties """

    def __init__(self, seq, start=0, header_lines=[], comments=[], end=[], load_headers=True, **kwargs):
        """
        An iteratable wrapper that coalesces headers and skips comments

        :param seq: An iterable
        :param start: The start of data row
        :param header_lines: An array of row numbers that should be coalesced into the header line, which is yieled first
        :param comments: An array of comment row numbers
        :param end: The last row number for data
        :param kwargs: Ignored. Sucks up extra parameters.
        :return:
        """

        self.iter = iter(seq)
        self.start = start if (start or start == 0) else 1
        self.header_lines = header_lines if isinstance(header_lines, (tuple, list)) else [int(e) for e in
                                                                                          header_lines.split(',') if e]
        self.comment_lines = comments
        self.end = end

        self.load_headers = load_headers

        self.headers = []
        self.comments = []

        int(self.start)  # Throw error if it is not an int

    @property
    def coalesce_headers(self):
        """Collects headers that are spread across multiple lines into a single row"""

        import re

        if not self.headers:
            return None

        header_lines = [list(hl) for hl in self.headers if bool(hl)]

        if len(header_lines) == 0:
            return []

        if len(header_lines) == 1:
            return header_lines[0]

        # If there are gaps in the values of a line, copy them forward, so there
        # is some value in every position
        for hl in header_lines:
            last = None
            for i in range(len(hl)):
                hli = str(hl[i])
                if not hli.strip():
                    hl[i] = last
                else:
                    last = hli

        headers = [' '.join(str(col_val).strip() if col_val else '' for col_val in col_set)
                   for col_set in zip(*header_lines)]

        headers = [re.sub(r'\s+', ' ', h.strip()) for h in headers]

        return headers

    def __iter__(self):

        row = []

        for i, row in enumerate(self.iter):

            if i in self.header_lines:
                if self.load_headers:
                    self.headers.append(row)
            elif i in self.comment_lines:
                self.comments.append(row)
            elif i == self.start:
                break

        if self.headers:
            headers = self.coalesce_headers
            yield headers
        else:
            # There is no header, so fake it

            headers = ['col' + str(i) for i, _ in enumerate(row)]
            yield headers

        yield row

        for row in self.iter:
            yield row


class ReorderRowGenerator(object):
    """A row generator that remaps columns. The row-generator must be in standard format,
    with the first row being the header, and all others being data.

     The column map maps dest_header->source_header ( it's inverted from what you might expect ). It can be
     a dict with dest values in keys and source values in values, or an object with `source' and 'dest' attributes,
     or a sequence of dicts with 'source' and 'dest' keys

     """

    def __init__(self, row_gen, colmap) -> None:

        self.row_gen = row_gen

        self.colmap = None

        if colmap:
            try:
                self.colmap = {e.dest: e.source for e in colmap}
            except AttributeError:
                pass

            if colmap and not self.colmap:
                try:
                    self.colmap = {e['dest']: e['source'] for e in colmap}
                except (KeyError, TypeError):
                    pass

            if colmap and not self.colmap:
                self.colmap = colmap

    def __iter__(self):

        itr = iter(self.row_gen)
        source_headers = next(itr)

        indexers = []

        if self.colmap:
            for dh, sh in self.colmap.items():
                try:
                    i = source_headers.index(sh)
                    indexers.append(f'row[{i}]')
                except ValueError:
                    indexers.append('None')

            code = 'lambda row: ({})'.format(','.join(indexers))

            f = eval(code)

            yield list(self.colmap.keys())
        else:

            f = lambda row: row

            yield source_headers

        for row in itr:
            yield f(row)



# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT, included in this distribution as LICENSE

""" """

from rowgenerators.appurl.web.web import WebUrl
from rowgenerators import parse_app_url

class S3Url(WebUrl):
    """Convert an S3 proto url into the public access form"""

    def __init__(self, url=None, downloader=None, **kwargs):
        # Save for auth_url()
        self._orig_url = url
        self._orig_kwargs = dict(kwargs.items())

        super().__init__(url,downloader=downloader, **kwargs)

    @classmethod
    def _match(cls, url, **kwargs):
        return url.proto == 's3';

    @property
    def auth_resource_url(self):
        """Return the orginal S3: version of the url, with a resource_url format that will trigger boto auth"""
        return 's3://{bucket}/{key}'.format(bucket=self.bucket_name, key=self.key)


    @property
    def resource_url(self):
        url_template = 'https://s3.amazonaws.com/{bucket}/{key}'
        return url_template.format(bucket=self.bucket_name, key=self.key)

    @property
    def resource_file(self):

        from rowgenerators.appurl import parse_app_url
        from rowgenerators.appurl.util import file_ext
        from os.path import basename, join, dirname

        return basename(self.resource_url)

    @property
    def resource_format(self):

        from rowgenerators.appurl import parse_app_url
        from rowgenerators.appurl.util import file_ext
        from os.path import basename, join, dirname

        if self._resource_format:
            return self._resource_format
        else:
            return file_ext(self.resource_file)

    def join_dir(self, s):

        from rowgenerators.appurl import parse_app_url
        from rowgenerators.appurl.util import file_ext
        from os.path import basename, join, dirname
        import pathlib

        try:
            path = s.path
        except AttributeError:
            path = parse_app_url(s).path

        # If there is a netloc, it's an absolute URL
        if s.netloc:
            return s

        new_key = str(pathlib.PurePosixPath(dirname(self.key)).joinpath(path))

        return parse_app_url('s3://{bucket}/{key}'.format(bucket=self.bucket_name.strip('/'), key=new_key.lstrip('/')))

    @property
    def bucket_name(self):
        return self.netloc

    @property
    def key(self):
        """S3 storage key, the file path"""
        return '' if not self.path else self.path.strip('/')

    @property
    def object(self):
        """Return the boto object for this source"""
        import boto3

        s3 = boto3.resource('s3')

        return s3.Object(self.bucket_name, self.key)

    def list(self):
        """List the top 'directory' of a S3 URL. Does not list recursively.  """
        import boto3
        import yaml
        client = boto3.client('s3')

        paginator = client.get_paginator('list_objects')

        for result in paginator.paginate(Bucket=self.bucket_name, Prefix=self.path.lstrip('/'), Delimiter='/'):

            if not result:
                continue

            # Contents, Name, Prefix, Delimiter, CommonPrefixes
            for e in result.get('Contents',[]):
                if e:
                    if e.get('Key') == result.get('Prefix'):
                        # The request was for a single file, not a prefix
                        url = "s3://" + self.bucket_name +result.get('Delimiter') + e.get('Key')
                    else:
                        url = "s3://"+self.bucket_name+result.get('Prefix')+result.get('Delimiter')+e.get('Key')

                    yield parse_app_url(url)

            for e in result.get('CommonPrefixes',[]):

                if e:
                    yield parse_app_url("s3://" + self.bucket_name +result.get('Delimiter') + e.get('Prefix'))

    def list_recursive(self):

        for e in self.list():

            if e.path.endswith('/'):
                yield from e.list_recursive()
            else:
                yield e


    @property
    def signed_resource_url(self):
        import boto3

        s3 = boto3.client('s3')

        url = s3.generate_presigned_url(
            ClientMethod='get_object',
            Params={
                'Bucket': self.bucket_name,
                'Key': self.key
            }
        )

        return url




# Copyright (c) 2016 Civic Knowledge. This file is licensed under the terms of the
# MIT License, included in this distribution as LICENSE.txt

"""

Built-in Transform functions.

"""


def info(v, row, row_n, i_s, i_d, header_s, header_d, scratch, errors, accumulator):
    """ Print information about a value, and return the value. Prints out these values:

    - row_n: The row number
    - header_d: Schema header
    - type: The python type of the value
    - value: The value of the row, truncated to 40 characters.

    :param v: The current value of the column
    :param row: A RowProxy object for the whiole row.
    :param row_n: The current row number.
    :param i_s: The numeric index of the source column
    :param i_d: The numeric index for the destination column
    :param header_s: The name of the source column
    :param header_d: The name of the destination column
    :param scratch: A dict that can be used for storing any values. Persists between rows.
    :param errors: A dict used to store error messages. Persists for all columns in a row, but not between rows.
    :param accumulator: A dict for use in accumulating values, such as computing aggregates.
    :return: The final value to be supplied for the column.
    """

    print("{}:{} {} {}".format(row_n, header_d, type(v), str(v)[:40]))

    return v



# -*- coding: utf-8 -*-
# Copyright (c) 2016 Civic Knowledge. This file is licensed under the terms of the
# MIT License, included in this distribution as LICENSE.txt

"""
A Row proxy provides a dict interface to row data, without having to re-construct a dict for each row.

NOTW: Don't store and reuse RowProxy objects; the iterators return the same RowProxy object, linked
to a different row each time, but with the same header. If you store them in a list, all of them
will have the row data of the last row. Comvert them to dicts ( RowProxy.dict ) before putting them into lists.

"""



class RowProxy(object):
    '''
    A dict-like accessor for rows which holds a constant header for the keys. Allows for faster access than
    constructing a dict, and also provides attribute access

    >>> header = list('abcde')
    >>> rp = RowProxy(header)
    >>> for i in range(10):
    >>>     row = [ j for j in range(len(header)]
    >>>     rp.set_row(row)
    >>>     print rp['c']

    '''

    def __init__(self, keys):

        self.__keys = keys
        self.__row = [None] * len(keys)
        self.__pos_map = {e: i for i, e in enumerate(keys)}
        self.__initialized = True

    @property
    def row(self):
        return object.__getattribute__(self, '_RowProxy__row')

    def set_row(self, v):
        object.__setattr__(self, '_RowProxy__row', v)
        return self

    @property
    def headers(self):
        return self.__getattribute__('_RowProxy__keys')

    def __setitem__(self, key, value):
        if isinstance(key, int):
            self.__row[key] = value
        else:
            self.__row[self.__pos_map[key]] = value

    def __getitem__(self, key):

            if isinstance(key, int):
                try:
                    return self.__row[key]
                except IndexError:
                    raise KeyError("Failed to get value for integer key {} in row {} ".format(key, self.__row))
            else:
                try:
                    return self.__row[self.__pos_map[key]]
                except IndexError:
                    raise IndexError("Failed to get value for non-int key '{}', resolved to position {} "
                                     .format(key, self.__pos_map[key]))
                except KeyError:
                    raise KeyError("Failed to get value for non-int key '{}'; has {} ".format(key, self.__keys))

    def get(self, item, default=None):

        try:
            return self.__getitem__(item)
        except (KeyError, IndexError):
            return default


    def __setattr__(self, key, value):

        if '_RowProxy__initialized' not in self.__dict__:
            return object.__setattr__(self, key, value)

        else:
            self.__row[self.__pos_map[key]] = value

    def __getattr__(self, key):
        try:
            return self.__row[self.__pos_map[key]]
        except KeyError:
            raise KeyError("Failed to find key '{}'; has {}".format(key, self.__keys))

    def __delitem__(self, key):
        raise NotImplementedError()

    def __iter__(self):
        return iter(self.__keys)

    def __len__(self):
        return len(self.__keys)

    @property
    def dict(self):
        return dict(zip(self.__keys, self.__row))


    def copy(self):
        return type(self)(self.__keys).set_row(list(self.row))

    def keys(self):
        return self.__keys

    def values(self):
        return self.__row

    def items(self):
        return zip(self.__keys, self.__row)

    # The final two methods aren't required, but nice for demo purposes:
    def __str__(self):
        """ Returns simple dict representation of the mapping. """
        return str(self.dict)

    def __repr__(self):
        return self.dict.__repr__()


class GeoRowProxy(RowProxy):

    @property
    def __geo_interface__(self):
        from shapely.wkt import loads

        g = loads(self.geometry)
        gi = g.__geo_interface__

        d = dict(self)
        del d['geometry']

        gi['properties'] = d

        return gi




"""Value Types

The value converters can recognize, parse, normalize and transform common codes, such as FIPS, ANSI and census codes.

Copyright (c) 2015 Civic Knowledge. This file is licensed under the terms of
the Revised BSD License, included in this distribution as LICENSE.txt

"""


from .types import *
from .core import *
from .dimensions import *
from .errors import *
from .measures import *
from .other import *
from .times import *
from .geo import *

value_types = {
    "int": IntMeasure,
    "long": LongMeasure,
    "float": FloatMeasure,
    "text": TextDimension,
    "str": StrDimension,
    "unknown": StrDimension,
    "unicode": TextDimension,
    "geometry": ShapeValue,
}

value_types.update(geo_value_types)
value_types.update(times_value_types)
value_types.update(dimension_value_types)
value_types.update(error_value_types)
value_types.update(measure_value_types)
value_types.update(other_value_types)



@memoize
def resolve_value_type(vt_code):
    import sys

    vt_code = vt_code.strip('?')

    try:

        o = value_types[vt_code]
        o.vt_code = vt_code
        return o
    except KeyError:

        parts = vt_code.split('/')
        args = []
        while len(parts):

            args.append(parts.pop())

            try:
                base_vt_code = '/'.join(parts)
                o = value_types[base_vt_code]
                # Return a dynamic subclass that has the extra parameters built in
                cls = o.subclass(vt_code, '/'.join(args))

                # Stuff the class into the current module
                globals()[cls.__name__] = cls

                # And also in the same module as the base class
                setattr(sys.modules[o.__module__], cls.__name__, cls)

                # FIXME. Shouldn't need this, but codegen.py seems to try to import
                # from the wrong module.
                setattr(sys.modules['rowgenerators.valuetype.core'], cls.__name__, cls)

                return cls

            except KeyError:
                pass



""" Value Type Exceptions

Functions for handling exceptions

"""


def clear_error(v):
    from rowgenerators.valuetype import FailedValue

    if isinstance(v, FailedValue):
        return None
    return v

def try_except( try_f, except_f):
    """Takes 2 closures and executes them in a try / except block """
    try:
        return try_f()
    except Exception as exception:
        return except_f(exception)

nan_value = float('nan')

def nan_is_none(v):
    import math

    try:
        if math.isnan(v):
            return None
        else:
            return v
    except (ValueError, TypeError):
        return v

def ignore(v):
    return None

def log_exception(exception, bundle):
    bundle.error(exception)
    return None




# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT License, included in this distribution as LICENSE.txt

""" """

from rowgenerators.source import Source
from rowgenerators.exceptions import RowGeneratorError
import pandas as pd
from functools import lru_cache

def bisect_df(df, v, col='variable_name_nd'):
    """Perform a binary search on a dataframe"""
    from math import log

    max_n = (log(len(df)) / log(2)) + 2

    low = 0
    high = len(df)
    mid = int((low + high) / 2)
    n = 0
    while True:

        e = df.iloc[mid]

        if v == e[col]:
            return e
        if v > e[col]:
            low = mid
        else:
            high = mid

        mid = int((low + high) / 2)

        if n > max_n:
            return None

        n += 1

        #print(n, max_n, v, low, mid, high)

    return None

@lru_cache()
def get_var_labels(ds_name, f):
    ds = f[ds_name+'_variable_labels']
    dhs = f[ds_name + '_variable_labels_headers']

    cols = dhs[:]
    return pd.DataFrame(ds[:], columns=cols)

class Hdf5Source(Source):
    """Generate rows from an excel file"""

    def __init__(self, ref, cache=None, working_dir=None, env=None, **kwargs):
        super().__init__(ref, cache, working_dir, **kwargs)

        self.url = ref

    def dataframe(self, limit=None, *args, **kwargs):
        import h5py
        import sys

        ds_name = self.ref.fragment[0]

        slice_frag = self.ref.fragment[1]

        with h5py.File(str(self.ref.fspath)) as f:

            ds = f[ds_name]

            try:
                if ':' in slice_frag:
                    parts = slice_frag.split(':')
                    slc = slice(int(parts[0]), int(parts[1]))
                else:
                    slc = [int(e) for e in slice_frag.split(',')]

                try:
                    headers_ds = f[ds_name + '_headers']
                    headers = list(headers_ds[slc])

                except KeyError:
                    headers = []

            except ValueError: # Hopefully b/c the slice is actually strings, not integers

                df = get_var_labels(ds_name, f)

                def name_to_num(v):
                    e = bisect_df(df,  v)
                    return int(e.col_no)

                if ':' in slice_frag:
                    slice_frag = slice_frag.split(':')
                    slc = slice(name_to_num(slice_frag[0]),name_to_num(slice_frag[1]))
                    headers = headers[slc]
                else:
                    slice_frag = slice_frag.split(',')
                    slc = [name_to_num(e) for e in slice_frag]
                    headers = slice_frag



            try:
                vars_ds = f[ds_name + '_variable_labels']
            except KeyError:
                headers = []

            headers_map = {i:e for i,e in enumerate(headers)}

            df = pd.DataFrame(ds[0:10,slc], columns=headers if headers else None)

        return df

    def __iter__(self):
        """Iterate over all of the lines in the file"""
        raise NotImplementedError()


# Copyright (c) 2016 Civic Knowledge. This file is licensed under the terms of the
# Revised BSD License, included in this distribution as LICENSE

from __future__ import print_function

import sys

from tabulate import tabulate

#Change the row cache name
from rowgenerators.util import  get_cache


def prt(*args):
    print(*args)

def warn( *args):
    print('WARN:',*args)

def err(*args):
    import sys
    print("ERROR:", *args)
    sys.exit(1)


def valuetypes():
    from rowgenerators.valuetype import value_types
    import argparse

    parser = argparse.ArgumentParser(
        prog='rowgen',
        description='Display information about rowgenerator valuetypes')

    parser.add_argument('-l', '--list', default=False, action='store_true',
                        help='List all of the valuetypes')

    cache = get_cache()

    args = parser.parse_args(sys.argv[1:])

    rows = [ (k,v.__name__, v.__doc__) for k,v in value_types.items() ]

    print(tabulate(sorted(rows), headers='Code Class Description'.split()))






from rowgenerators.appurl.file import FileUrl


class StataUrl(FileUrl):
    """
    """
    match_priority = 85

    @classmethod
    def _match(cls, url, **kwargs):
        return url.proto == 'file' and url.target_format == 'dta'


    def __init__(self, url=None, downloader=None, **kwargs):
        super().__init__(url, downloader, **kwargs)









# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT, included in this distribution as LICENSE

"""
URLS for referencing python code.
"""

import sys
from .file import FileUrl
from rowgenerators.appurl.util import import_name_or_class

class PythonUrl(FileUrl):
    """
    URL to reference python code
    """

    match_priority = FileUrl.match_priority - 1

    def __init__(self, url=None, downloader=None, **kwargs):

        super().__init__(url, downloader, **kwargs)
        kwargs['scheme'] = 'python'

    @classmethod
    def _match(cls, url, **kwargs):
        return url.proto == 'python' or url.resource_format == 'py'

    def get_resource(self):
        return self

    def get_target(self):
        return self

    @property
    def object(self):
        """Return the python thing, a class or a function, that will be invoked """

        try:
            path = self.path.replace('/', '.') + '.' + self.target_file
            return import_name_or_class(path)

        except ModuleNotFoundError:
            pass


        if self.fspath.parent not in sys.path:
            sys.path.insert(0, str(self.fspath.parent))

        mod = __import__(self.fspath.stem)

        return getattr(mod,self.target_file)


    def __call__(self, *args, **kwargs):

        try:
            return self.object(*args, **kwargs)
        except TypeError as e:
            raise TypeError(str(e)+"; for args='{}', kwargs='{}' ".format(args, kwargs))



